"""Client to download ERA5-Land data products from the climate data store.

See <https://cds.climate.copernicus.eu/datasets/reanalysis-era5-land?tab=overview>.
"""

from __future__ import annotations

import importlib.resources
import json
import logging
import tempfile
import zipfile
from calendar import monthrange
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from functools import cached_property
from math import ceil
from pathlib import Path
from time import sleep
from typing import Iterator

import geopandas as gpd
import xarray as xr
from datapi import ApiClient, Remote
from requests.exceptions import HTTPError

with importlib.resources.open_text("openhexa.toolbox.era5", "variables.json") as f:
    VARIABLES = json.load(f)

DATASET = "reanalysis-era5-land"

log = logging.getLogger(__name__)


@dataclass
class DataRequest:
    """CDS data request as expected by the API."""

    variable: list[str]
    year: str
    month: str
    day: list[str]
    time: list[str]
    data_format: str = "grib"
    area: list[float] | None = None


def bounds_from_file(fp: Path, buffer: float = 0.5) -> list[float]:
    """Get bounds from file.

    Parameters
    ----------
    fp : Path
        File path.
    buffer : float, optional
        Buffer to add to the bounds (default=0.5).

    Returns
    -------
    list[float]
        Bounds (north, west, south, east).
    """
    boundaries = gpd.read_parquet(fp)
    xmin, ymin, xmax, ymax = boundaries.total_bounds
    xmin = ceil(xmin - buffer)
    ymin = ceil(ymin - buffer)
    xmax = ceil(xmax + buffer)
    ymax = ceil(ymax + buffer)
    return ymax, xmin, ymin, xmax


def get_period_chunk(dtimes: list[datetime]) -> dict:
    """Get the period chunk for a list of datetimes.

    The period chunk is a dictionary with the "year", "month", "day" and "time" keys as expected by
    the CDS API. A period chunk cannot contain more than 1 year and 1 month. However, it can
    contain any number of days and times.

    This is the temporal part of a CDS data request.

    Parameters
    ----------
    dtimes : list[datetime]
        A list of datetimes for which we want data

    Returns
    -------
    dict
        The period chunk, in other words the temporal part of the request payload

    Raises
    ------
    ValueError
        If the list of datetimes contains more than 1 year or more than 1 month
    """
    years = {dtime.year for dtime in dtimes}
    if len(years) > 1:
        msg = "Cannot create a period chunk for multiple years"
        raise ValueError(msg)
    months = {dtime.month for dtime in dtimes}
    if len(months) > 1:
        msg = "Cannot create a period chunk for multiple months"
        raise ValueError(msg)

    year = next(iter(years))
    month = next(iter(months))
    days = []

    for dtime in sorted(dtimes):
        if dtime.day not in days:
            days.append(dtime.day)

    return {
        "year": str(year),
        "month": f"{month:02}",
        "day": [f"{day:02}" for day in days],
    }


def iter_chunks(dtimes: list[datetime]) -> Iterator[dict]:
    """Get the period chunks for a list of datetimes.

    The period chunks are a list of dictionaries with the "year", "month", "day" and "time" keys as
    expected by the CDS API. A period chunk cannot contain more than 1 year and 1 month. However,
    it can contain any number of days and times.

    The function tries its best to generate the minimum amount of chunks to minimize the amount of requests.

    Parameters
    ----------
    dtimes : list[datetime]
        A list of datetimes for which we want data

    Returns
    -------
    Iterator[dict]
        The period chunks (one per month max)
    """
    for year in range(min(dtimes).year, max(dtimes).year + 1):
        for month in range(12):
            dtimes_month = [dtime for dtime in dtimes if dtime.year == year and dtime.month == month + 1]
            if dtimes_month:
                yield get_period_chunk(dtimes_month)


def list_datetimes_in_dataset(ds: xr.Dataset) -> list[datetime]:
    """List datetimes in input dataset for which data is available.

    It is assumed that the dataset has a `time` dimension, in addition to `latitude` and `longitude`
    dimensions. We consider that a datetime is available in a dataset if non-null data values are
    present for more than 1 step.
    """
    dtimes = []
    data_vars = list(ds.data_vars)
    var = data_vars[0]

    for time in ds.time.values:
        dtime = datetime.fromtimestamp(time.astype(int) / 1e9, tz=timezone.utc)
        if dtime in dtimes:
            continue
        non_null = ds.sel(time=time)[var].notnull().sum().values.item()
        if non_null > 0:
            dtimes.append(dtime)

    return dtimes


def list_datetimes_in_dir(data_dir: Path) -> list[datetime]:
    """List datetimes in datasets that can be found in an input directory."""
    dtimes = []

    for f in data_dir.glob("*.grib"):
        # sometimes gribs are actually zip files
        with tempfile.NamedTemporaryFile(mode="wb") as tmp:
            if zipfile.is_zipfile(f):
                with zipfile.ZipFile(f, "r") as zip:
                    tmp.write(zip.read("data.grib"))
                ds = xr.open_dataset(tmp.name, engine="cfgrib")

            else:
                ds = xr.open_dataset(f, engine="cfgrib")

            dtimes += list_datetimes_in_dataset(ds)

    dtimes = sorted(set(dtimes))

    msg = f"Scanned {data_dir.as_posix()}, found data for {len(dtimes)} dates"
    log.info(msg)

    return dtimes


def date_range(start: datetime, end: datetime) -> list[datetime]:
    """Get a range of dates with a 1-day step."""
    drange = []
    dt = start
    while dt <= end:
        drange.append(dt)
        dt += timedelta(days=1)
    return drange


def build_request(
    variable: str,
    year: int,
    month: int,
    day: list[int] | list[str] | None = None,
    time: list[int] | list[str] | None = None,
    data_format: str = "grib",
    area: list[float] | None = None,
) -> DataRequest:
    """Build request payload.

    Parameters
    ----------
    variable : str
        Climate data store variable name (ex: "2m_temperature").
    year : int
        Year of interest.
    month : int
        Month of interest.
    day : list[int] | list[str] | None, optional
        Days of interest. Defauls to None (all days).
    time : list[int] | list[str] | None, optional
        Hours of interest (ex: [1, 6, 18]). Defaults to None (all hours).
    data_format : str, optional
        Output data format ("grib" or "netcdf"). Defaults to "grib".
    area : list[float] | None, optional
        Area of interest (north, west, south, east). Defaults to None (world).

    Returns
    -------
    DataRequest
        CDS data equest payload.

    Raises
    ------
    ValueError
        Request parameters are not valid.
    """
    if variable not in VARIABLES:
        msg = f"Variable {variable} not supported"
        raise ValueError(msg)

    if data_format not in ["grib", "netcdf"]:
        msg = f"Data format {data_format} not supported"
        raise ValueError(msg)

    # in the CDS data request, area is an array of float or int in the following order:
    # [north, west, south, east]
    if area:
        n, w, s, e = area
        msg = "Invalid area of interest"
        max_lat = 90
        max_lon = 180
        if ((abs(n) > max_lat) or (abs(s) > max_lat)) or ((abs(w) > max_lon) or (abs(e) > max_lon)):
            raise ValueError(msg)
        if (n < s) or (e < w):
            raise ValueError(msg)

    # in the CDS data request, days must be an array of strings (one string per day)
    # ex: ["01", "02", "03"]
    if not day:
        dmax = monthrange(year, month)[1]
        day = list(range(1, dmax + 1))

    if isinstance(day[0], int):
        day = [f"{d:02}" for d in day]

    # in the CDS data request, time must be an array of strings (one string per hour)
    # only hours between 00:00 and 23:00 are valid
    # ex: ["00:00", "03:00", "06:00"]
    if not time:
        time = range(24)

    if isinstance(time[0], int):
        time = [f"{hour:02}:00" for hour in time]

    return DataRequest(
        variable=[variable],
        year=str(year),
        month=f"{month:02}",
        day=day,
        time=time,
        data_format="grib",
        area=list(area) if area else None,
    )


class CDS:
    """Climate data store API client based on datapi."""

    def __init__(self, key: str, url: str = "https://cds.climate.copernicus.eu/api") -> None:
        """Initialize CDS client."""
        self.client = ApiClient(key=key, url=url)
        self.client.check_authentication()
        msg = f"Sucessfully authenticated to {url}"
        log.info(msg)

    @cached_property
    def latest(self) -> datetime:
        """Get date of latest available product."""
        collection = self.client.get_collection(DATASET)
        return collection.end_datetime

    def get_remote_requests(self) -> list[dict]:
        """Fetch list of the last 100 data requests in the CDS account."""
        requests = []
        jobs = self.client.get_jobs(limit=100)
        for request_id in jobs.request_uids:
            try:
                remote = self.client.get_remote(request_id)
                if remote.status in ["failed", "dismissed", "deleted"]:
                    continue
                requests.append({"request_id": request_id, "request": remote.request})
            except HTTPError:
                continue
        return requests

    def get_remote_from_request(self, request: DataRequest, existing_requests: list[dict]) -> Remote | None:
        """Look for a remote object that matches the provided request payload.

        Parameters
        ----------
        request : DataRequest
            Data request payload to look for.
        existing_requests : list[dict]
            List of existing data requests (as returned by self.get_remote_requests()).

        Returns
        -------
        Remote | None
            Remote object if found, None otherwise.
        """
        if not existing_requests:
            return None

        for remote_request in existing_requests:
            if remote_request["request"] == request.__dict__:
                return self.client.get_remote(remote_request["request_id"])

        return None

    def submit(self, request: DataRequest) -> Remote:
        """Submit an async data request to the CDS API.

        If an identical data request has already been submitted, the Remote object corresponding to
        the existing data request is returned instead of submitting a new one.
        """
        return self.client.submit(DATASET, request=request.__dict__)

    def retrieve(self, request: DataRequest, dst_file: Path | str) -> None:
        """Submit and download a data request to the CDS API."""
        dst_file = Path(dst_file)
        dst_file.parent.mkdir(parents=True, exist_ok=True)
        self.client.retrieve(collection_id=DATASET, target=dst_file, request=request.__dict__)

    def download_between(
        self,
        start: datetime,
        end: datetime,
        variable: str,
        area: list[float],
        dst_dir: str | Path,
        time: list[int] | None = None,
    ) -> None:
        """Download all ERA5 data files needed to cover the period.

        Data requests are sent asynchronously (max one per month) to the CDS API and fetched when
        they are completed.

        Parameters
        ----------
        start : datetime
            Start date.
        end : datetime
            End date.
        variable : str
            Climate data store variable name (ex: "2m_temperature").
        area : list[float]
            Area of interest (north, west, south, east).
        dst_dir : str | Path
            Output directory.
        time : list[int] | None, optional
            Hours of interest (ex: [1, 6, 18]). Defaults to None (all hours).
        """
        dst_dir = Path(dst_dir)
        dst_dir.mkdir(parents=True, exist_ok=True)

        if not start.tzinfo:
            start = start.astimezone(tz=timezone.utc)
        if not end.tzinfo:
            end = end.astimezone(tz=timezone.utc)

        if end > self.latest:
            end = self.latest
            msg = "End date is after latest available product, setting end date to {}".format(end.strftime("%Y-%m-%d"))
            log.info(msg)

        # get the list of dates for which we will want to download data, which is the difference
        # between the available (already downloaded) and the requested dates
        drange = date_range(start, end)
        available = [dtime.date() for dtime in list_datetimes_in_dir(dst_dir)]
        dates = [d for d in drange if d.date() not in available]
        msg = f"Will request data for {len(dates)} dates"
        log.info(msg)

        existing_requests = self.get_remote_requests()
        remotes: list[Remote] = []

        for chunk in iter_chunks(dates):
            request = build_request(variable=variable, data_format="grib", area=area, time=time, **chunk)

            # has a similar request been submitted recently? if yes, use it instead of submitting
            # a new one
            remote = self.get_remote_from_request(request, existing_requests)
            if remote:
                remotes.append(remote)
                msg = f"Found existing request for date {request.year}-{request.month}"
                log.info(msg)
            else:
                remote = self.submit(request)
                remotes.append(remote)
                msg = f"Submitted new data request {remote.request_id} for {request.year}-{request.month}"

        while remotes:
            for remote in remotes:
                if remote.results_ready:
                    request = remote.request
                    fname = f"{request['year']}{request['month']}_{remote.request_id}.grib"
                    dst_file = Path(dst_dir, fname)
                    remote.download(dst_file.as_posix())
                    msg = f"Downloaded {dst_file.name}"
                    log.info(msg)
                    remotes.remove(remote)
                    remote.delete()

            if remotes:
                msg = f"Still {len(remotes)} files to download. Waiting 30s before retrying..."
                log.info(msg)
                sleep(30)



from __future__ import annotations

import calendar
import datetime
import logging
import re
from dataclasses import dataclass
from typing import Generator, Self
from warnings import warn

from dateutil import relativedelta

from openhexa.toolbox.dhis2.api import DHIS2ToolboxError

logger = logging.getLogger(__name__)


class InvalidPeriodError(DHIS2ToolboxError):
    """Exception raised for invalid period formats."""

    pass


class MixedPeriodsError(DHIS2ToolboxError):
    """Exception raised for mixed period types."""

    pass


class Period:
    """Base class for all period types."""

    def range(self, other: Self) -> Generator[Self]:
        """Generate all periods between self and other, inclusive."""
        if not isinstance(other, type(self)):
            msg = "Range can only be generated between two Period objects of the same type."
            raise InvalidPeriodError(msg)

        start_date = self.start
        end_date = other.start

        if start_date >= end_date:
            raise InvalidPeriodError("Start date must be before end date.")

        current_date = start_date
        while current_date <= end_date:
            yield type(self).from_date(current_date)
            current_date += self.delta

    def get_range(self, other: Self) -> list[Self]:
        """Get a range of periods between self and other, inclusive.

        DEPRECATED: Use the range method instead.
        """
        warn(
            "Period.get_range is deprecated. Use Period.range instead.",
            DeprecationWarning,
        )
        return list(self.range(other))

    @property
    def _delta(self) -> datetime.timedelta | relativedelta.relativedelta:
        """Return the delta for the period.

        DEPRECATED: Use Period.delta instead.
        """
        warn(
            "Period._delta is deprecated. Use Period.delta instead.",
            DeprecationWarning,
        )
        return self.delta

    @property
    def datetime(self) -> datetime.datetime | datetime.date:
        """Return the datetime representation of the period.

        DEPRECATED: Use Period.start or Period.end instead.
        """
        warn(
            "Period.datetime is deprecated. Use Period.start or Period.end instead.",
            DeprecationWarning,
        )
        return self.start

    @property
    def period(self) -> str:
        """Return the string representation of the period.

        DEPRECATED: Use str(Period) instead.
        """
        warn(
            "Period.period is deprecated. Use str(Period) instead.",
            DeprecationWarning,
        )
        return str(self)


@dataclass
class Day(Period):
    """Represents a single day period.

    Format: yyyyMMdd
    Example: 20040315
    Description: March 15, 2004

    Attributes
    ----------
    year : int
        The year of the period.
    month : int
        The month of the period.
    day : int
        The day of the period.
    """

    year: int
    month: int
    day: int
    delta: datetime.timedelta = datetime.timedelta(days=1)

    def __str__(self):
        return f"{self.year:04}{self.month:02}{self.day:02}"

    @property
    def start(self) -> datetime.date:
        return datetime.date(self.year, self.month, self.day)

    @property
    def end(self) -> datetime.date:
        return datetime.date(self.year, self.month, self.day)

    @classmethod
    def from_date(cls, dt: datetime.datetime | datetime.date):
        return cls(dt.year, dt.month, dt.day)

    @classmethod
    def from_string(cls, date_str: str):
        if not re.match(_PERIODS_REGEX[Day], date_str):
            raise InvalidPeriodError(f"Invalid period string: {date_str}")
        year = int(date_str[:4])
        month = int(date_str[4:6])
        day = int(date_str[6:])
        return cls(year, month, day)


@dataclass
class Week(Period):
    """Represents a week period.

    Format: yyyyWn
    Example: 2004W10
    Description: Week 10 2004

    Attributes
    ----------
    year : int
        The year of the period.
    week : int
        The week of the period.
    """

    year: int
    week: int
    delta: relativedelta.relativedelta = relativedelta.relativedelta(weeks=1)

    def __str__(self):
        return f"{self.year:04}W{self.week}"

    def __post_init__(self):
        if self.week < 1 or self.week > 53:
            msg = f"Invalid week number: {self.week}"
            raise InvalidPeriodError(msg)

        if self.year < 1900:
            msg = f"Invalid year: {self.year}"
            raise InvalidPeriodError(msg)

    @classmethod
    def from_date(cls, date: datetime.date | datetime.datetime):
        return cls(year=date.isocalendar()[0], week=date.isocalendar()[1])

    @classmethod
    def from_string(cls, week_str: str) -> Self:
        if not re.match(_PERIODS_REGEX[Week], week_str):
            raise InvalidPeriodError(f"Invalid period string: {week_str}")
        year, week = map(int, week_str.split("W"))
        return cls(year, week)

    @property
    def start(self) -> datetime.date:
        return datetime.date.fromisocalendar(self.year, self.week, 1)

    @property
    def end(self) -> datetime.date:
        return datetime.date.fromisocalendar(self.year, self.week, 7)


class WeekWednesday(Week):
    """Represents a week period starting on Wednesday.

    Format: yyyyWedWn
    Example: 2015WedW5
    Description: Week 5 with start Wednesday

    Attributes
    ----------
    year : int
        The year of the period.
    week : int
        The week of the period.
    """

    def __str__(self):
        return f"{self.year:04}WedW{self.week}"

    @classmethod
    def from_date(cls, date: datetime.date | datetime.datetime):
        adjusted_date = date - datetime.timedelta(days=2)
        return cls(year=adjusted_date.isocalendar()[0], week=adjusted_date.isocalendar()[1])

    @classmethod
    def from_string(cls, week_str: str) -> Self:
        if not re.match(_PERIODS_REGEX[WeekWednesday], week_str):
            raise InvalidPeriodError(f"Invalid period string: {week_str}")
        year, week = map(int, week_str.split("WedW"))
        return cls(year, week)

    @property
    def start(self) -> datetime.date:
        return datetime.date.fromisocalendar(self.year, self.week, 1) + datetime.timedelta(days=2)

    @property
    def end(self) -> datetime.date:
        return datetime.date.fromisocalendar(self.year, self.week, 7) + datetime.timedelta(days=2)


class WeekThursday(Week):
    """Represents a week period starting on Thursday.

    Format: yyyyThuWn
    Example: 2015ThuW6
    Description: Week 6 with start Thursday

    Attributes
    ----------
    year : int
        The year of the period.
    week : int
        The week of the period.
    """

    def __str__(self):
        return f"{self.year:04}ThuW{self.week}"

    @classmethod
    def from_date(cls, date: datetime.date | datetime.datetime):
        adjusted_date = date - datetime.timedelta(days=3)
        return cls(year=adjusted_date.isocalendar()[0], week=adjusted_date.isocalendar()[1])

    @classmethod
    def from_string(cls, week_str: str) -> Self:
        if not re.match(_PERIODS_REGEX[WeekThursday], week_str):
            raise InvalidPeriodError(f"Invalid period string: {week_str}")
        year, week = map(int, week_str.split("ThuW"))
        return cls(year, week)

    @property
    def start(self) -> datetime.date:
        return datetime.date.fromisocalendar(self.year, self.week, 1) + datetime.timedelta(days=3)

    @property
    def end(self) -> datetime.date:
        return datetime.date.fromisocalendar(self.year, self.week, 7) + datetime.timedelta(days=3)


class WeekSaturday(Week):
    """Represents a week period starting on Saturday.

    Format: yyyySatWn
    Example: 2015SatW7
    Description: Week 7 with start Saturday

    Attributes
    ----------
    year : int
        The year of the period.
    week : int
        The week of the period.
    """

    def __str__(self):
        return f"{self.year:04}SatW{self.week}"

    @classmethod
    def from_date(cls, date: datetime.date | datetime.datetime):
        adjusted_date = date + datetime.timedelta(days=2)
        return cls(year=adjusted_date.isocalendar()[0], week=adjusted_date.isocalendar()[1])

    @classmethod
    def from_string(cls, week_str: str) -> Self:
        if not re.match(_PERIODS_REGEX[WeekSaturday], week_str):
            raise InvalidPeriodError(f"Invalid period string: {week_str}")
        year, week = map(int, week_str.split("SatW"))
        return cls(year, week)

    @property
    def start(self) -> datetime.date:
        return datetime.date.fromisocalendar(self.year, self.week, 1) - datetime.timedelta(days=2)

    @property
    def end(self) -> datetime.date:
        return datetime.date.fromisocalendar(self.year, self.week, 7) - datetime.timedelta(days=2)


class WeekSunday(Week):
    """Represents a week period starting on Sunday.

    Format: yyyySunWn
    Example: 2015SunW8
    Description: Week 8 with start Sunday

    Attributes
    ----------
    year : int
        The year of the period.
    week : int
        The week of the period.
    """

    def __str__(self):
        return f"{self.year:04}SunW{self.week}"

    @classmethod
    def from_date(cls, date: datetime.date | datetime.datetime):
        adjusted_date = date + datetime.timedelta(days=1)
        return cls(year=adjusted_date.isocalendar()[0], week=adjusted_date.isocalendar()[1])

    @classmethod
    def from_string(cls, week_str: str) -> Self:
        if not re.match(_PERIODS_REGEX[WeekSunday], week_str):
            raise InvalidPeriodError(f"Invalid period string: {week_str}")
        year, week = map(int, week_str.split("SunW"))
        return cls(year, week)

    @property
    def start(self) -> datetime.date:
        return datetime.date.fromisocalendar(self.year, self.week, 1) - datetime.timedelta(days=1)

    @property
    def end(self) -> datetime.date:
        return datetime.date.fromisocalendar(self.year, self.week, 7) - datetime.timedelta(days=1)


@dataclass
class BiWeek(Period):
    """Represents a bi-week period.

    Format: yyyyBiWn
    Example: 2015BiW1
    Description: Week 1-2 2015

    Attributes
    ----------
    year : int
        The year of the period.
    bi_week : int
        The bi-week of the period.
    """

    year: int
    bi_week: int
    delta: relativedelta.relativedelta = relativedelta.relativedelta(weeks=2)

    def __str__(self):
        return f"{self.year:04}BiW{self.bi_week}"

    def __post_init__(self):
        if self.bi_week < 1 or self.bi_week > 27:
            msg = f"Invalid biweek number: {self.bi_week}"
            raise InvalidPeriodError(msg)

        if self.year < 1900:
            msg = f"Invalid year: {self.year}"
            raise InvalidPeriodError(msg)

    @classmethod
    def from_date(cls, date: datetime.date | datetime.datetime):
        week = Week.from_date(date)
        bi_week = (week.week - 1) // 2 + 1
        return cls(year=week.year, bi_week=bi_week)

    @classmethod
    def from_string(cls, bi_week_str: str):
        if not re.match(_PERIODS_REGEX[BiWeek], bi_week_str):
            raise InvalidPeriodError(f"Invalid period string: {bi_week_str}")
        year, bi_week = map(int, bi_week_str.split("BiW"))
        return cls(year, bi_week)

    @property
    def start(self) -> datetime.date:
        week = (self.bi_week - 1) * 2 + 1
        return datetime.date.fromisocalendar(self.year, week, 1)

    @property
    def end(self) -> datetime.date:
        week = self.bi_week * 2
        return datetime.date.fromisocalendar(self.year, week, 7)


@dataclass
class Month(Period):
    """Represents a month period.

    Format: yyyyMM
    Example: 200403
    Description: March 2004

    Attributes
    ----------
    year : int
        The year of the period.
    month : int
        The month of the period.
    """

    year: int
    month: int
    delta: relativedelta.relativedelta = relativedelta.relativedelta(months=1)

    def __str__(self):
        return f"{self.year:04}{self.month:02}"

    def __post_init__(self):
        if self.month < 1 or self.month > 12:
            msg = f"Invalid month: {self.month}"
            raise InvalidPeriodError(msg)

        if self.year < 1900:
            msg = f"Invalid year: {self.year}"
            raise InvalidPeriodError(msg)

    @classmethod
    def from_date(cls, date: datetime.date | datetime.datetime):
        return cls(year=date.year, month=date.month)

    @classmethod
    def from_string(cls, month_str: str):
        if not re.match(_PERIODS_REGEX[Month], month_str):
            raise InvalidPeriodError(f"Invalid period string: {month_str}")
        year = int(month_str[:4])
        month = int(month_str[4:])
        return cls(year, month)

    @property
    def start(self) -> datetime.date:
        return datetime.date(self.year, self.month, 1)

    @property
    def end(self) -> datetime.date:
        last_day_of_month = calendar.monthrange(self.year, self.month)[1]
        return datetime.date(self.year, self.month, last_day_of_month)


@dataclass
class BiMonth(Period):
    """Represents a bi-month period.

    Format: yyyyMMB
    Example: 200401B
    Description: January-February 2004

    Attributes
    ----------
    year : int
        The year of the period.
    bi_month : int
        The bi-month of the period.
    """

    year: int
    bi_month: int
    delta: relativedelta.relativedelta = relativedelta.relativedelta(months=2)

    def __str__(self):
        return f"{self.year:04}{self.bi_month:02}B"

    def __post_init__(self):
        if self.bi_month < 1 or self.bi_month % 2 == 0:
            msg = f"Invalid bi-month: {self.bi_month}"
            raise InvalidPeriodError(msg)

        if self.year < 1900:
            msg = f"Invalid year: {self.year}"
            raise InvalidPeriodError(msg)

    @classmethod
    def from_date(cls, date: datetime.date | datetime.datetime):
        month = Month.from_date(date)
        bi_month = (month.month - 1) // 2 * 2 + 1
        return cls(year=month.year, bi_month=bi_month)

    @classmethod
    def from_string(cls, bi_month_str: str):
        if not re.match(_PERIODS_REGEX[BiMonth], bi_month_str):
            raise InvalidPeriodError(f"Invalid period string: {bi_month_str}")
        year = int(bi_month_str[:4])
        bi_month = int(bi_month_str[4:6])
        return cls(year, bi_month)

    @property
    def start(self) -> datetime.date:
        return datetime.date(self.year, self.bi_month, 1)

    @property
    def end(self) -> datetime.date:
        last_day_of_month = calendar.monthrange(self.year, self.bi_month + 1)[1]
        return datetime.date(self.year, self.bi_month + 1, last_day_of_month)


@dataclass
class Quarter(Period):
    """Represents a quarter period.

    Format: yyyyQn
    Example: 2004Q1
    Description: January-March 2004

    Attributes
    ----------
    year : int
        The year of the period.
    quarter : int
        The quarter of the period.
    """

    year: int
    quarter: int
    delta: relativedelta.relativedelta = relativedelta.relativedelta(months=3)

    def __str__(self):
        return f"{self.year:04}Q{self.quarter}"

    def __post_init__(self):
        if not (1 <= self.quarter <= 5):
            msg = f"Invalid quarter: {self.quarter}"
            raise InvalidPeriodError(msg)

        if self.year < 1900:
            msg = f"Invalid year: {self.year}"
            raise InvalidPeriodError(msg)

    @classmethod
    def from_date(cls, date: datetime.date | datetime.datetime):
        month = Month.from_date(date)
        quarter = (month.month - 1) // 3 + 1
        return cls(year=month.year, quarter=quarter)

    @classmethod
    def from_string(cls, quarter_str: str):
        if not re.match(_PERIODS_REGEX[Quarter], quarter_str):
            raise InvalidPeriodError(f"Invalid period string: {quarter_str}")
        year, quarter = map(int, quarter_str.split("Q"))
        return cls(year, quarter)

    @property
    def start(self) -> datetime.date:
        month = (self.quarter - 1) * 3 + 1
        return datetime.date(self.year, month, 1)

    @property
    def end(self) -> datetime.date:
        month = self.quarter * 3
        last_day_of_month = calendar.monthrange(self.year, month)[1]
        return datetime.date(self.year, month, last_day_of_month)


@dataclass
class SixMonth(Period):
    """Represents a six-month period.

    Format: yyyySn
    Example: 2004S1
    Description: January-June 2004

    Attributes
    ----------
    year : int
        The year of the period.
    six_month : int
        The six-month of the period.
    """

    year: int
    six_month: int
    delta: relativedelta.relativedelta = relativedelta.relativedelta(months=6)

    def __str__(self):
        return f"{self.year:04}S{self.six_month}"

    def __post_init__(self):
        if self.six_month not in {1, 2}:
            msg = f"Invalid six-month: {self.six_month}"
            raise InvalidPeriodError(msg)

        if self.year < 1900:
            msg = f"Invalid year: {self.year}"
            raise InvalidPeriodError(msg)

    @classmethod
    def from_date(cls, date: datetime.date | datetime.datetime):
        six_month = 1 if date.month <= 6 else 2
        return cls(year=date.year, six_month=six_month)

    @classmethod
    def from_string(cls, six_month_str: str):
        if not re.match(_PERIODS_REGEX[SixMonth], six_month_str):
            raise InvalidPeriodError(f"Invalid period string: {six_month_str}")
        year, six_month = map(int, six_month_str.split("S"))
        return cls(year, six_month)

    @property
    def start(self) -> datetime.date:
        month = (self.six_month - 1) * 6 + 1
        return datetime.date(self.year, month, 1)

    @property
    def end(self) -> datetime.date:
        month = self.six_month * 6
        last_day_of_month = calendar.monthrange(self.year, month)[1]
        return datetime.date(self.year, month, last_day_of_month)


@dataclass
class Year(Period):
    """Represents a year period.

    Format: yyyy
    Example: 2004
    Description: 2004

    Attributes
    ----------
    year : int
        The year of the period.
    """

    year: int
    delta: relativedelta.relativedelta = relativedelta.relativedelta(years=1)

    def __str__(self):
        return f"{self.year:04}"

    def __post_init__(self):
        if self.year < 1900:
            msg = f"Invalid year: {self.year}"
            raise InvalidPeriodError(msg)

    @classmethod
    def from_date(cls, date: datetime.date | datetime.datetime):
        return cls(year=date.year)

    @classmethod
    def from_string(cls, year_str: str):
        if not re.match(_PERIODS_REGEX[Year], year_str):
            raise InvalidPeriodError(f"Invalid period string: {year_str}")
        year = int(year_str)
        return cls(year)

    @property
    def start(self) -> datetime.date:
        return datetime.date(self.year, 1, 1)

    @property
    def end(self) -> datetime.date:
        return datetime.date(self.year, 12, 31)


@dataclass
class FinancialApril(Period):
    """Represents a financial year period starting in April.

    Format: yyyyApril
    Example: 2004April
    Description: April 2004-March 2005

    Attributes
    ----------
    year : int
        The year of the period.
    """

    year: int
    delta: relativedelta.relativedelta = relativedelta.relativedelta(years=1)

    def __str__(self):
        return f"{self.year:04}April"

    def __post_init__(self):
        if self.year < 1900:
            msg = f"Invalid year: {self.year}"
            raise InvalidPeriodError(msg)

    @classmethod
    def from_date(cls, date: datetime.date | datetime.datetime):
        return cls(year=date.year)

    @classmethod
    def from_string(cls, year_str: str):
        if not re.match(_PERIODS_REGEX[FinancialApril], year_str):
            raise InvalidPeriodError(f"Invalid period string: {year_str}")
        year = int(year_str[:4])
        return cls(year)

    @property
    def start(self) -> datetime.date:
        return datetime.date(self.year, 4, 1)

    @property
    def end(self) -> datetime.date:
        return datetime.date(self.year + 1, 3, 31)


@dataclass
class FinancialJuly(Period):
    """Represents a financial year period starting in July.

    Format: yyyyJuly
    Example: 2004July
    Description: July 2004-June 2005

    Attributes
    ----------
    year : int
        The year of the period.
    """

    year: int
    delta: relativedelta.relativedelta = relativedelta.relativedelta(years=1)

    def __str__(self):
        return f"{self.year:04}July"

    def __post_init__(self):
        if self.year < 1900:
            msg = f"Invalid year: {self.year}"
            raise InvalidPeriodError(msg)

    @classmethod
    def from_date(cls, date: datetime.date | datetime.datetime):
        return cls(year=date.year)

    @classmethod
    def from_string(cls, year_str: str):
        if not re.match(_PERIODS_REGEX[FinancialJuly], year_str):
            raise InvalidPeriodError(f"Invalid period string: {year_str}")
        year = int(year_str[:4])
        return cls(year)

    @property
    def start(self) -> datetime.date:
        return datetime.date(self.year, 7, 1)

    @property
    def end(self) -> datetime.date:
        return datetime.date(self.year + 1, 6, 30)


@dataclass
class FinancialOct(Period):
    """Represents a financial year period starting in October.

    Format: yyyyOct
    Example: 2004Oct
    Description: October 2004-September 2005

    Attributes
    ----------
    year : int
        The year of the period.
    """

    year: int
    delta: relativedelta.relativedelta = relativedelta.relativedelta(years=1)

    def __str__(self):
        return f"{self.year:04}Oct"

    def __post_init__(self):
        if self.year < 1900:
            msg = f"Invalid year: {self.year}"
            raise InvalidPeriodError(msg)

    @classmethod
    def from_date(cls, date: datetime.date | datetime.datetime):
        return cls(year=date.year)

    @classmethod
    def from_string(cls, year_str: str):
        if not re.match(_PERIODS_REGEX[FinancialOct], year_str):
            raise InvalidPeriodError(f"Invalid period string: {year_str}")
        year = int(year_str[:4])
        return cls(year)

    @property
    def start(self) -> datetime.date:
        return datetime.date(self.year, 10, 1)

    @property
    def end(self) -> datetime.date:
        return datetime.date(self.year + 1, 9, 30)


@dataclass
class FinancialNov(Period):
    """Represents a financial year period starting in November.

    Format: yyyyNov
    Example: 2004Nov
    Description: November 2004-October 2005

    Attributes
    ----------
    year : int
        The year of the period.
    """

    year: int
    delta: relativedelta.relativedelta = relativedelta.relativedelta(years=1)

    def __str__(self):
        return f"{self.year:04}Nov"

    def __post_init__(self):
        if self.year < 1900:
            msg = f"Invalid year: {self.year}"
            raise InvalidPeriodError(msg)

    @classmethod
    def from_date(cls, date: datetime.date | datetime.datetime):
        return cls(year=date.year)

    @classmethod
    def from_string(cls, year_str: str):
        if not re.match(_PERIODS_REGEX[FinancialNov], year_str):
            raise InvalidPeriodError(f"Invalid period string: {year_str}")
        year = int(year_str[:4])
        return cls(year)

    @property
    def start(self) -> datetime.date:
        return datetime.date(self.year, 11, 1)

    @property
    def end(self) -> datetime.date:
        return datetime.date(self.year + 1, 10, 31)


_PERIODS_REGEX = {
    Day: r"^\d{4}\d{2}\d{2}$",
    Week: r"^\d{4}W([1-9]\d*)$",
    Month: r"^\d{4}\d{2}$",
    BiMonth: r"^\d{4}\d{2}B$",
    Quarter: r"^\d{4}Q\d$",
    SixMonth: r"^\d{4}S\d$",
    Year: r"^\d{4}$",
    FinancialApril: r"^\d{4}April$",
    FinancialJuly: r"^\d{4}July$",
    FinancialOct: r"^\d{4}Oct$",
    FinancialNov: r"^\d{4}Nov$",
    WeekWednesday: r"^\d{4}WedW([1-9]\d*)$",
    WeekThursday: r"^\d{4}ThuW([1-9]\d*)$",
    WeekSaturday: r"^\d{4}SatW([1-9]\d*)$",
    WeekSunday: r"^\d{4}SunW([1-9]\d*)$",
    BiWeek: r"^\d{4}BiW([1-9]\d*)$",
}


def period_from_string(period_str: str) -> Period:
    """Convert a string to a Period object.

    Parameters
    ----------
    period_str : str
        The period string to convert.

    Returns
    -------
    Period
        The corresponding Period object.

    Raises
    ------
    InvalidPeriodError
        If the period string is not valid.
    """
    for period_type, regex in _PERIODS_REGEX.items():
        if re.match(regex, period_str):
            return period_type.from_string(period_str)

    raise InvalidPeriodError(f"Invalid period string: {period_str}")


def get_range(start: Period, end: Period) -> list[Period]:
    """Get a range of periods between start and end, inclusive.

    DEPRECATED: Use Period.range instead.
    """
    warn(
        "get_range is deprecated. Use Period.range instead.",
        DeprecationWarning,
    )
    if not isinstance(start, type(end)):
        raise MixedPeriodsError("Start and end periods must be of the same type.")

    return list(start.range(end))



"""Download raw historical Era5 products from Google Cloud:
https://console.cloud.google.com/storage/browser/gcp-public-data-arco-era5

Products are provided as raw NetCDF files and are usually available with a ~3 month lag.
"""

from __future__ import annotations

import importlib.resources
import json
import logging
import shutil
import tempfile
from datetime import datetime, timedelta
from functools import cached_property
from pathlib import Path

import requests
from google.cloud import storage

with importlib.resources.open_text("openhexa.toolbox.era5", "variables.json") as f:
    VARIABLES = json.load(f)

log = logging.getLogger(__name__)


class NotFoundError(Exception):
    pass


class ParameterError(ValueError):
    pass


class Client:
    def __init__(self):
        self.client = storage.Client.create_anonymous_client()
        self.bucket = self.client.bucket("gcp-public-data-arco-era5")

    @staticmethod
    def prefix(variable: str, date: datetime) -> str:
        """Build key prefix for a given product."""
        return f"raw/date-variable-single_level/{date.year}/{date.month:02}/{date.day:02}/{variable}/surface.nc"

    def _subdirs(self, prefix: str) -> list[str]:
        """List subdirs."""
        blobs = self.client.list_blobs(self.bucket, prefix=prefix, delimiter="/")
        prefixes = []
        for page in blobs.pages:
            prefixes += page.prefixes
        return prefixes

    @cached_property
    def latest(self) -> datetime:
        """Get date of latest available product."""
        root = "raw/date-variable-single_level/"
        subdirs = self._subdirs(root)  # years
        subdirs = self._subdirs(max(subdirs))  # months
        subdirs = self._subdirs(max(subdirs))  # days
        subdir = max(subdirs).split("/")
        year = int(subdir[-4])
        month = int(subdir[-3])
        day = int(subdir[-2])
        return datetime(year, month, day)

    def find(self, variable: str, date: datetime) -> str | None:
        """Find public URL of product. Return None if not found."""
        prefix = self.prefix(variable, date)
        blobs = self.client.list_blobs(self.bucket, prefix=prefix, max_results=1)
        blobs = list(blobs)
        if blobs:
            return blobs[0].public_url
        else:
            return None

    def download(self, variable: str, date: datetime, dst_file: str | Path, overwrite=False):
        """Download an Era5 NetCDF product for a given day.

        Parameters
        ----------
        variable : str
            Climate data store variable name (ex: "2m_temperature").
        date : datetime
            Product date (year, month, day).
        dst_file : str | Path
            Output file.
        overwrite : bool, optional
            Overwrite existing file (default=False).

        Raises
        ------
        ParameterError
            Product request parameters are invalid.
        NotFoundError
            Product not found in bucket.
        """
        dst_file = Path(dst_file)
        dst_file.parent.mkdir(parents=True, exist_ok=True)

        if dst_file.exists() and not overwrite:
            log.debug("Skipping download of %s because file already exists", str(dst_file.absolute()))
            return

        if variable not in VARIABLES:
            raise ParameterError("%s is not a valid climate data store variable name", variable)

        url = self.find(variable, date)
        if not url:
            raise NotFoundError("%s product not found for date %s", variable, date.strftime("%Y-%m-%d"))

        with tempfile.NamedTemporaryFile() as tmp:
            with open(tmp.name, "wb") as f:
                with requests.get(url, stream=True) as r:
                    for chunk in r.iter_content(chunk_size=1024**2):
                        if chunk:
                            f.write(chunk)

            shutil.copy(tmp.name, dst_file)

        log.debug("Downloaded %s", str(dst_file.absolute()))

    def sync(self, variable: str, start_date: datetime, end_date: datetime, dst_dir: str | Path):
        """Download all products for a given variable and date range.

        If products are already present in the destination directory, they will be skipped.
        Expects file names to be formatted as "YYYY-MM-DD_VARIABLE.nc".

        Parameters
        ----------
        variable : str
            Climate data store variable name (ex: "2m_temperature").
        start_date : datetime
            Start date (year, month, day).
        end_date : datetime
            End date (year, month, day).
        dst_dir : str | Path
            Output directory.
        """
        dst_dir = Path(dst_dir)
        dst_dir.mkdir(parents=True, exist_ok=True)

        if start_date > end_date:
            raise ParameterError("`start_date` must be before `end_date`")

        date = start_date
        if end_date > self.latest:
            log.info("Setting `end_date` to the latest available date: %s" % date.strftime("%Y-%m-%d"))
            end_date = self.latest

        while date <= end_date:
            expected_filename = f"{date.strftime('%Y-%m-%d')}_{variable}.nc"
            fpath = Path(dst_dir, expected_filename)
            fpath_grib = Path(dst_dir, expected_filename.replace(".nc", ".grib"))
            if fpath.exists() or fpath_grib.exists():
                log.debug("%s already exists, skipping download" % expected_filename)
            else:
                self.download(variable=variable, date=date, dst_file=fpath, overwrite=False)
            date += timedelta(days=1)



import io
import typing

from openhexa.toolbox.iaso.api_client import ApiClient
import polars as pl


class IASO:
    """
    The IASO toolbox provides an interface to interact with the IASO.
    """

    def __init__(self, url: str, username: str, password: str) -> None:
        """
        Initializes the IASO toolbox.
        :param url: IASO server URL
        :param username: IASO instance username
        :param password: IASO instance password

        Examples:
            >>> from openhexa.toolbox.iaso import IASO
            >>> iaso = IASO(server_url="http://iaso-staging.bluesquare.org",
            >>>             username="user",
            >>>             password="pass")
        """
        self.api_client = ApiClient(url, username, password)

    def get_projects(self, page: int = 1, limit: int = 10, **kwargs) -> dict:
        """
        Fetches projects list from IASO. Method is paginated by default. Pagination can be modified and additional
        arguments can be passed as key value parameters.

        Examples:
            >>> from openhexa.toolbox.iaso import IASO
            >>> iaso = IASO(client=ApiClient(url="http://iaso-staging.bluesquare.org",
            >>>             username="user",
            >>>             password="pass"))
            >>> iaso.get_projects(page=1, limit=1, id=1)
        """

        params = kwargs
        params.update({"page": page, "limit": limit})
        response = self.api_client.get("/api/projects", params=params)
        return response.json().get("projects")

    def get_org_units(self, page: int = 1, limit: int = 10, optimized=False, **kwargs) -> dict:
        """
        Fetches org units from IASO. Method is paginated by default. Pagination can be modified and additional
        arguments can be passed as key value parameters.

        Examples:
            >>> from openhexa.toolbox.iaso import IASO
            >>> iaso = IASO(client=ApiClient(url="http://iaso-staging.bluesquare.org",
            >>>             username="user",
            >>>             password="pass"))
            >>> projects = iaso.get_org_units(page=1, limit=1, id=1)
        """
        org_units_endpoint = "/api/orgunits"
        params = kwargs
        params.update({"page": page, "limit": limit})
        params.update({"validation_status": "VALID"})

        if optimized:
            org_units_endpoint = "/api/orgunits/tree/search"
            if "search" in kwargs:
                params.update({"smallSearch": "true"})

        response = self.api_client.get(org_units_endpoint, params=params)
        json_response = response.json()
        if optimized and "results" in json_response:
            return json_response.get("results")
        elif "orgunits" in json_response:
            return json_response.get("orgunits")
        elif "orgUnits" in json_response:
            return json_response.get("orgUnits")
        return {}

    def get_form_instances(
        self,
        page: int = 1,
        limit: int = 10,
        as_dataframe: bool = False,
        **kwargs,
    ) -> typing.Union[dict, pl.DataFrame]:
        """
        Fetches form instances from IASO filtered by form id. Method is paginated by default.
        Pagination can be modified and additional arguments can be passed as key value parameters.
        There is a possiblity to fetch forms as DataFrames.

        Params:
            :param page: The page number of the form instance.
            :param limit: The maximum number of form instances.
            :param as_dataframe: If true, will return a DataFrame containing form instances.
            :param kwargs: additonal arguments passed to the /forms endpoint as URL parameters.

        Examples:
            >>> from openhexa.toolbox.iaso import IASO
            >>> iaso = IASO(url="http://iaso-staging.bluesquare.org", username="user", password="pass")
            >>> form_dataframes = iaso.get_form_instances(page=1, limit=1, as_dataframe=True, ids=276)
        """

        params = kwargs
        params.update({"page": page, "limit": limit})
        params.update({"page": page, "limit": limit})
        if as_dataframe:
            params.update({"csv": "true"})
            response = self.api_client.get("/api/instances", params=params)
            forms = pl.read_csv(io.StringIO(response.content.decode("utf-8")))
            return forms
        response = self.api_client.get("/api/instances/", params=kwargs)
        forms = response.json().get("instances")
        return forms

    def get_forms(
        self,
        page: int = 1,
        limit: int = 10,
        **kwargs,
    ) -> dict:
        """
        Fetches forms from IASO. Method is paginated by default.
        Pagination can be modified and additional arguments can be passed as key value parameters.

        Params:
            :param page: The page number of the form.
            :param limit: The maximum number of form.
            :param kwargs: additonal arguments passed to the /forms endpoint as URL parameters.

        Examples:
            >>> from openhexa.toolbox.iaso import IASO
            >>> iaso = IASO(url="http://iaso-staging.bluesquare.org",username="user",password="pass")
            >>> forms_by_orgunits_and_projects = iaso.get_forms(page=1, limit=1, org_units=[300], projects=[23])
            >>> forms_by_orgunits_and_projects = iaso.get_forms(page=1, limit=1, search="test")
        """
        params = kwargs
        params.update({"page": page, "limit": limit})
        response = self.api_client.get("/api/forms", params=params)
        return response.json().get("forms")



"""A set of opinionated functions to extract data and metadata from a IASO account into dataframes."""

from __future__ import annotations

import json
from datetime import datetime
from io import BytesIO, StringIO
from typing import Iterable

import fiona
import polars as pl

from openhexa.toolbox.iaso import IASO


def _get_org_units_csv(iaso: IASO) -> str:
    """Extract org units in CSV format."""
    r = iaso.api_client.get(url="api/orgunits", params={"csv": True}, stream=True, timeout=30)
    r.raise_for_status()
    return r.content.decode("utf8")


def _get_org_units_gpkg(iaso: IASO) -> bytes:
    """Extract org units in GPKG format."""
    r = iaso.api_client.get(url="api/orgunits", params={"gpkg": True}, stream=True, timeout=30)
    r.raise_for_status()
    return r.content


def _get_org_units_geometries(iaso: IASO) -> dict[int, str]:
    """Get the org units geometries from IASO.

    Org unit geometries are absent from the CSV export, so we need to fetch them separately.

    Parameters
    ----------
    iaso: IASO
        The IASO client.

    Returns
    -------
    dict[int, str]
        A dict with org unit ids as keys and GeoJSON geometries as values.
    """
    gpkg = _get_org_units_gpkg(iaso)
    features = {}
    layers = fiona.listlayers(BytesIO(gpkg))
    for layer in layers:
        with fiona.open(BytesIO(gpkg), layer=layer) as src:
            for feature in src:
                if not feature.geometry:
                    continue
                ou_id = int(feature["properties"]["id"])
                geom = json.dumps(feature["geometry"].__geo_interface__)
                features[ou_id] = geom
    return features


def get_organisation_units(iaso: IASO) -> pl.DataFrame:
    """Get the organisation units from IASO.

    Parameters
    ----------
    iaso: IASO
        The IASO client.

    Returns
    -------
    pl.DataFrame
        The organisation units dataframe.
    """
    csv = _get_org_units_csv(iaso)
    df = pl.read_csv(StringIO(csv))

    df = df.select(
        pl.col("ID").alias("id"),
        pl.col("Nom").alias("name"),
        pl.col("Type").alias("org_unit_type"),
        pl.col("Latitude").alias("latitude"),
        pl.col("Longitude").alias("longitude"),
        pl.col("Date d'ouverture").str.to_date("%Y-%m-%d").alias("opening_date"),
        pl.col("Date de fermeture").str.to_date("%Y-%m-%d").alias("closing_date"),
        pl.col("Date de création").str.to_datetime("%Y-%m-%d %H:%M").alias("created_at"),
        pl.col("Date de modification").str.to_datetime("%Y-%m-%d %H:%M").alias("updated_at"),
        pl.col("Source").alias("source"),
        pl.col("Validé").alias("validation_status"),
        pl.col("Référence externe").alias("source_ref"),
        *[
            pl.col(f"Ref Ext parent {lvl}").alias(f"level_{lvl}_ref")
            for lvl in range(1, 10)
            if f"Ref Ext parent {lvl}" in df.columns
        ],
        *[pl.col(f"parent {lvl}").alias(f"level_{lvl}_name") for lvl in range(1, 10) if f"parent {lvl}" in df.columns],
    )

    geoms = _get_org_units_geometries(iaso)
    df = df.with_columns(
        pl.col("id").map_elements(lambda x: geoms.get(x, None), return_dtype=pl.String).alias("geometry")
    )

    return df


def _iter_children(children: list[dict]) -> Iterable[dict]:
    for child in children:
        if child.get("type"):
            yield child
            if child.get("children"):
                yield from _iter_children(child["children"])


def _get_form_versions(iaso: IASO, form_id: int) -> dict:
    """Extract form versions metadata from IASO."""
    r = iaso.api_client.get(url="api/formversions", params={"form_id": form_id, "fields": "descriptor"}, timeout=5)
    r.raise_for_status()
    return r.json()


def _get_questions(descriptor: dict) -> dict:
    """Extract questions metadata from form descriptor.

    Parameters
    ----------
    descriptor: dict
        The form descriptor.

    Returns
    -------
    dict
        A dict with question names as keys and metadata as values.
    """
    questions = {}
    for child in _iter_children(descriptor["children"]):
        questions[child["name"]] = {
            "name": child["name"],
            "type": child["type"],
            "label": child.get("label"),
            "list_name": child.get("list_name"),
            "calculate": child["bind"].get("calculate") if child["type"] == "calculate" else None,
        }
    return questions


def _get_choices(descriptor: dict) -> dict:
    """Extract choices metadata from form descriptor.

    Parameters
    ----------
    descriptor: dict
        The form descriptor.

    Returns
    -------
    dict
        A dict with choice list names as keys and a list of choices as values.
    """
    all_choices = {}
    for child in _iter_children(descriptor["children"]):
        if child["type"].startswith("select"):
            question_choices = child["children"]
            list_name = child["list_name"]
            if list_name not in all_choices:
                all_choices[list_name] = []
            for choice in question_choices:
                if choice not in all_choices[list_name]:
                    all_choices[list_name].append(choice)
    return all_choices


def get_form_metadata(iaso: IASO, form_id: int) -> dict:
    """Get form metadata from IASO.

    Return a dict with form versions as keys and, for each version, two dicts:
      * questions: a dict with metadata for each question, with question name as key
        includes name, type, label, list_name and calculate expression
      * choices: a dict with metadata for each choice list, with list_name as key
        includes choice name and label for each choice

    Parameters
    ----------
    iaso: IASO
        The IASO client.
    form_id: int
        The form id.

    Returns
    -------
    dict
        A dict with form versions as keys and, for each version, questions and choices metadata.
    """
    form_versions = _get_form_versions(iaso=iaso, form_id=form_id)
    meta = {}

    for version in form_versions["form_versions"]:
        descriptor = version["descriptor"]
        ver = int(descriptor["version"])
        meta[ver] = {}
        meta[ver]["questions"] = _get_questions(descriptor=descriptor)
        meta[ver]["choices"] = _get_choices(descriptor=descriptor)

    return meta


def _get_instances_csv(iaso: IASO, form_id: int, last_updated: str | None = None) -> str:
    """Extract form instances in CSV format."""
    params = {"form_id": form_id, "csv": True}
    if last_updated is not None:
        params["modificationDateFrom"] = last_updated
    r = iaso.api_client.get(url="api/instances", params=params, stream=True, timeout=30)
    r.raise_for_status()
    return r.content.decode("utf8")


def _process_instance(instance: dict, form_metadata: dict, mapping: dict) -> dict:
    """Create a dict row from a submission instance.

    Also handles casting of values based on ODK question type.

    Parameters
    ----------
    instance: dict
        The submission instance.
    form_metadata: dict
        The form metadata for all versions.
    mapping: dict
        Mapping between ODK question types and Polars data types.

    Returns
    -------
    dict
        The row as a dict (with column names as key).
    """
    row = {
        "id": instance.get("ID du formulaire"),
        "form_version": instance.get("Version du formulaire"),
        "created_at": datetime.strptime(
            instance.get("Date de création"),
            "%Y-%m-%d %H:%M:%S",
        ),
        "updated_at": datetime.strptime(
            instance.get("Date de modification"),
            "%Y-%m-%d %H:%M:%S",
        ),
        "org_unit_id": instance.get("Org unit id"),
        "org_unit_name": instance.get("Org unit"),
        "org_unit_ref": instance.get("Référence externe"),
        "latitude": instance.get("Latitude"),
        "longitude": instance.get("Longitude"),
    }

    form_version = instance.get("Version du formulaire")
    questions = form_metadata[form_version]["questions"]

    for name, question in questions.items():
        type = question["type"]
        if type not in mapping:
            continue

        src_value = instance.get(name)
        if src_value == "" or src_value is None:
            dst_value = None

        elif type == "integer":
            dst_value = int(src_value)
        elif type == "decimal":
            dst_value = float(src_value)
        elif type == "select all that apply":
            dst_value = [str(v) for v in src_value.split(" ")]
        elif type == "date":
            dst_value = datetime.strptime(src_value, "%Y-%m-%d").date()
        elif type == "time":
            dst_value = datetime.strptime(src_value, "%H:%M:%S.%f%z").time()
        elif type == "datetime":
            dst_value = datetime.strptime(src_value, "%Y-%m-%dT%H:%M:%S%z")
        elif type == "geopoint":
            coords = [float(v) for v in src_value.split(" ")]
            feature = {"type": "Point", "coordinates": [coords[1], coords[0]]}
            dst_value = json.dumps(feature)
        elif type == "rank":
            dst_value = [str(v) for v in src_value.split(" ")]
        elif type == "repeat":
            dst_value = json.dumps(src_value)
        else:
            dst_value = str(src_value)

        row[name] = dst_value

    return row


def _merge_schemas(schemas: list[dict]) -> dict:
    """Merge multiple schemas into one.

    Merge data types of questions with the same name. If data types differ between form versions,
    then the most permissive type is used (e.g. if one version is int and another is str,
    then the merged type is str).

    Parameters
    ----------
    schemas: list[dict]
        A list of schemas to merge.

    Returns
    -------
    dict
        The merged schema.
    """
    merged_schema = {}

    for schema in schemas:
        for column in schema:
            if column not in merged_schema:
                merged_schema[column] = []  # use a list to store all types
            merged_schema[column].append(schema[column])

    final_schema = {}
    for column, dtypes in merged_schema.items():
        dtype = dtypes[0]
        for dtype_ in dtypes[1:]:
            if dtype_ != dtype:
                dtype = pl.String  # use string as the most permissive type
        final_schema[column] = dtype

    return final_schema


def extract_submissions(iaso: IASO, form_id: int, last_updated: str | None = None) -> pl.DataFrame:
    """Extract submissions instances for a IASO form.

    Parameters
    ----------
    iaso: IASO
        The IASO client.
    form_id: int
        The form id.
    last_updated: str, optional
        The last updated date to fetch in ISO format.

    Returns
    -------
    pl.DataFrame
        The submissions dataframe with one row per submission.
    """
    csv = _get_instances_csv(iaso=iaso, form_id=form_id, last_updated=last_updated)
    form_metadata = get_form_metadata(iaso=iaso, form_id=form_id)
    rows = []

    # Polars schema for default instance properties
    base_schema = {
        "id": pl.String,
        "form_version": pl.String,
        "created_at": pl.Datetime(time_unit="us", time_zone=None),
        "updated_at": pl.Datetime(time_unit="us", time_zone=None),
        "org_unit_id": pl.Int64,
        "org_unit_name": pl.String,
        "org_unit_ref": pl.String,
        "latitude": pl.Float64,
        "longitude": pl.Float64,
    }

    # mapping between ODK question types and target Polars data types
    mapping = {
        "text": pl.String,
        "integer": pl.Int64,
        "decimal": pl.Float64,
        "select one": pl.String,
        "select all that apply": pl.List(pl.String),
        "date": pl.Date,
        "time": pl.Time,
        "datetime": pl.Datetime,
        "geopoint": pl.String,
        "photo": pl.String,
        "file": pl.String,
        "audio": pl.String,
        "video": pl.String,
        "barcode": pl.String,
        "acknowledge": pl.String,
        "calculate": pl.String,
        "repeat": pl.String,
        "rank": pl.List(pl.String),
        "range": pl.String,
    }

    instances = pl.read_csv(StringIO(csv))

    for instance in instances.iter_rows(named=True):
        row = _process_instance(instance=instance, form_metadata=form_metadata, mapping=mapping)
        rows.append(row)

    # build polars schemas for all form versions that are found in the data
    versions_with_submissions = instances["Version du formulaire"].unique().to_list()
    schemas = []
    for form_version, form_meta in form_metadata.items():
        if form_version not in versions_with_submissions:
            continue
        schema = base_schema.copy()
        for name, question in form_meta["questions"].items():
            if question["type"] not in mapping:
                continue
            schema[name] = mapping[question["type"]]
        schemas.append(schema)

    # schemas are merged - if multiple data types are found for the same question across versions,
    # the most permissive type is used (e.g. str)
    schema = _merge_schemas(schemas=schemas)

    return pl.DataFrame(rows, schema=schema)


def replace_labels(submissions: pl.DataFrame, form_metadata: dict, language: str | None = None) -> pl.DataFrame:
    """Replace choice list values with labels.

    Parameters
    ----------
    submissions: pl.DataFrame
        The submissions dataframe.
    form_metadata: dict
        The form metadata for all versions.
    questions: dict
        The questions metadata.
    choices: dict
        The choices metadata.
    language: str, optional
        The language code for the labels. Must be specified for multi-language forms.

    Returns
    -------
    pl.DataFrame
        The submissions dataframe with choice values replaced by labels.
    """
    # build label mapping for all form versions
    # keep "select one", "select all that apply" and "rank" types separate because they
    # need to be handled differently
    mapping = {}
    for form_version, form_meta in form_metadata.items():
        mapping[form_version] = {"select one": {}, "select all that apply": {}, "rank": {}}
        questions = form_meta["questions"]
        choices = form_meta["choices"]
        for name, question in questions.items():
            qtype = question["type"]
            if qtype not in ["select one", "select all that apply", "rank"]:
                continue
            if question not in submissions.columns:
                continue
            mapping[form_version][qtype][name] = {}
            for choice in choices[question["list_name"]]:
                # in single-language forms, choice labels are strings
                if isinstance(choice["label"], str):
                    mapping[form_version][qtype][name][choice["name"]] = choice["label"]
                # in multi-language forms, choice labels are dicts with language as keys
                else:
                    if language not in choice["label"]:
                        raise ValueError(f"Language {language} not found in choice list labels")
                    mapping[form_version][qtype][name][choice["name"]] = choice["label"].get(language)

    # replace values with labels (select one)
    for form_version, label_mapping in mapping.items():
        form_version = str(form_version)
        for question, choices in label_mapping["select one"].items():
            submissions = submissions.with_columns(
                pl.when(pl.col("form_version") == form_version)
                .then(pl.col(question).map_elements(lambda x: choices.get(x, x), return_dtype=pl.String))
                .otherwise(pl.col(question))
                .alias(question)
            )
        for question, choices in label_mapping["select all that apply"].items():
            submissions = submissions.with_columns(
                pl.when(pl.col("form_version") == form_version)
                .then(
                    pl.col(question).map_elements(lambda x: [choices.get(v, v) for v in x], return_dtype=pl.List(str))
                )
                .otherwise(pl.col(question))
                .alias(question)
            )
        for question, choices in label_mapping["rank"].items():
            submissions = submissions.with_columns(
                pl.when(pl.col("form_version") == form_version)
                .then(
                    pl.col(question).map_elements(lambda x: [choices.get(v, v) for v in x], return_dtype=pl.List(str))
                )
                .otherwise(pl.col(question))
                .alias(question)
            )

    return submissions



"""Module for spatial and temporal aggregation of ERA5 data."""

from datetime import datetime
from pathlib import Path

import geopandas as gpd
import numpy as np
import polars as pl
import rasterio
import xarray as xr
from epiweeks import Week
from rasterio.features import rasterize
from rasterio.transform import Affine, from_bounds


def clip_dataset(ds: xr.Dataset, xmin: float, ymin: float, xmax: float, ymax: float) -> xr.Dataset:
    """Clip input xarray dataset according to the provided bounding box.

    Assumes lat & lon dimensions are named "latitude" and "longitude". Longitude in the
    source dataset is expected to be in the range [0, 360], and will be converted to
    [-180, 180].

    Parameters
    ----------
    ds : xr.Dataset
        Input xarray dataset.
    xmin : float
        Minimum longitude.
    ymin : float
        Minimum latitude.
    xmax : float
        Maximum longitude.
    ymax : float
        Maximum latitude.

    Returns
    -------
    xr.Dataset
        Clipped xarray dataset.
    """
    ds = ds.assign_coords(longitude=(((ds.longitude + 180) % 360) - 180)).sortby("longitude")
    ds = ds.where((ds.longitude >= xmin) & (ds.longitude <= xmax), drop=True)
    ds = ds.where((ds.latitude >= ymin) & (ds.latitude <= ymax), drop=True)
    return ds


def get_transform(ds: xr.Dataset) -> Affine:
    """Get rasterio affine transform from xarray dataset.

    Parameters
    ----------
    ds : xr.Dataset
        Input xarray dataset.

    Returns
    -------
    Affine
        Rasterio affine transform.
    """
    transform = from_bounds(
        ds.longitude.values.min(),
        ds.latitude.values.min(),
        ds.longitude.values.max(),
        ds.latitude.values.max(),
        len(ds.longitude),
        len(ds.latitude),
    )
    return transform


def build_masks(
    boundaries: gpd.GeoDataFrame, height: int, width: int, transform: rasterio.Affine
) -> tuple[np.ndarray, rasterio.Affine]:
    """Build binary masks for all geometries in a dataframe.

    We build a raster of shape (n_boundaries, n_height, n_width) in order to store one binary mask
    per boundary. Boundaries shapes cannot be stored in a single array as we want masks to overlap
    if needed.

    Parameters
    ----------
    boundaries : gpd.GeoDataFrame
        Input GeoDataFrame containing the boundaries.
    height : int
        Height of the raster (number of pixels)
    width : int
        Width of the raster (number of pixels)
    transform : rasterio.Affine
        Raster affine transform

    Returns
    -------
    np.ndarray
        Binary masks as a numpy ndarray of shape (n_boundaries, height, width)
    """
    masks = np.ndarray(shape=(len(boundaries), height, width), dtype=np.bool_)
    for i, geom in enumerate(boundaries.geometry):
        mask = rasterize(
            shapes=[geom.__geo_interface__],
            out_shape=(height, width),
            fill=0,
            default_value=1,
            all_touched=True,
            transform=transform,
        )
        masks[i, :, :] = mask == 1
    return masks


def merge(data_dir: Path | str) -> xr.Dataset:
    """Merge all .grib files in a directory into a single xarray dataset.

    If multiple values are available for a given time, step, longitude & latitude dimensions, the
    maximum value is kept.

    Parameters
    ----------
    data_dir : Path | str
        Directory containing the .grib files.

    Returns
    -------
    xr.Dataset
        Merged xarray dataset with time, step, longitude and latitude dimensions.
    """
    if isinstance(data_dir, str):
        data_dir = Path(data_dir)

    files = data_dir.glob("*.grib")
    ds = xr.open_dataset(next(files), engine="cfgrib")

    for f in files:
        ds = xr.concat([ds, xr.open_dataset(f, engine="cfgrib")], dim="tmp_dim").max(dim="tmp_dim")

    return ds


def _np_to_datetime(dt64: np.datetime64) -> datetime:
    epoch = np.datetime64(0, "s")
    one_second = np.timedelta64(1, "s")
    seconds_since_epoch = (dt64 - epoch) / one_second
    return datetime.fromtimestamp(seconds_since_epoch)


def _has_missing_data(da: xr.DataArray) -> bool:
    """A DataArray is considered to have missing data if not all hours have measurements."""
    missing = False

    # if da.step.size == 1, da.step is just an int so we cannot iterate over it
    # if da.step size > 1, da.step is an array of int (one per step)
    if da.step.size > 1:
        for step in da.step:
            if da.sel(step=step).isnull().all():
                missing = True
    else:
        missing = da.isnull().all()

    return missing


def _week(date: datetime) -> str:
    year = date.isocalendar()[0]
    week = date.isocalendar()[1]
    return f"{year}W{week}"


def _epi_week(date: datetime) -> str:
    epiweek = Week.fromdate(date)
    year = epiweek.year
    week = epiweek.week
    return f"{year}W{week}"


def _month(date: datetime) -> str:
    return date.strftime("%Y%m")


def aggregate(ds: xr.Dataset, var: str, masks: np.ndarray, boundaries_id: list[str]) -> pl.DataFrame:
    """Aggregate hourly measurements in space and time.

    Parameters
    ----------
    ds : xr.Dataset
        Input xarray dataset with time, step, longitude and latitude dimensions
    var : str
        Variable to aggregate (ex: "t2m" or "tp")
    masks : np.ndarray
        Binary masks as a numpy ndarray of shape (n_boundaries, height, width)
    boundaries_id : list[str]
        List of boundary IDs (same order as n_boundaries dimension in masks)

    Notes
    -----
    The function aggregates hourly measurements to daily values for each boundary.

    Temporal aggregation is applied first. 3 statistics are computed for each day: daily mean,
    daily min, and daily max.

    Spatial aggregation is then applied. For each boundary, 3 statistics are computed: average of
    daily means, average of daily min, and average of daily max. These 3 statistics are stored in
    the "mean", "min", and "max" columns of the output dataframe.
    """
    rows = []

    for day in ds.time.values:
        da = ds[var].sel(time=day)

        if _has_missing_data(da):
            continue

        # if there is a step dimension (= hourly measurements), aggregate to daily
        # if not, data is already daily
        if "step" in da.dims:
            da_mean = da.mean(dim="step").values
            da_min = da.min(dim="step").values
            da_max = da.max(dim="step").values
        else:
            da_mean = da.values
            da_min = da.values
            da_max = da.values

        for i, uid in enumerate(boundaries_id):
            v_mean = np.nanmean(da_mean[masks[i, :, :]])
            v_min = np.nanmin(da_min[masks[i, :, :]])
            v_max = np.nanmax(da_max[masks[i, :, :]])

            rows.append(
                {
                    "boundary_id": uid,
                    "date": _np_to_datetime(day).date(),
                    "mean": v_mean,
                    "min": v_min,
                    "max": v_max,
                }
            )

    SCHEMA = {
        "boundary_id": pl.String,
        "date": pl.Date,
        "mean": pl.Float64,
        "min": pl.Float64,
        "max": pl.Float64,
    }

    df = pl.DataFrame(data=rows, schema=SCHEMA)

    # add week, month, and epi_week period columns
    df = df.with_columns(
        pl.col("date").map_elements(_week, str).alias("week"),
        pl.col("date").map_elements(_month, str).alias("month"),
        pl.col("date").map_elements(_epi_week, str).alias("epi_week"),
    )

    return df


def aggregate_per_week(
    daily: pl.DataFrame,
    column_uid: str,
    use_epidemiological_weeks: bool = False,
    sum_aggregation: bool = False,
) -> pl.DataFrame:
    """Aggregate daily data per week.

    Parameters
    ----------
    daily : pl.DataFrame
        Daily data with a "week" or "epi_week", "mean", "min", and "max" columns
        Length of the dataframe should be (n_boundaries * n_days).
    column_uid : str
        Column containing the boundary ID.
    use_epidemiological_weeks : bool, optional
        Use epidemiological weeks instead of iso weeks.
    sum_aggregation : bool, optional
        If True, sum values instead of computing the mean, for example for total precipitation data.

    Returns
    -------
    pl.DataFrame
        Weekly aggregated data of length (n_boundaries * n_weeks).
    """
    if use_epidemiological_weeks:
        week_column = "epi_week"
    else:
        week_column = "week"

    df = daily.select([column_uid, pl.col(week_column).alias("week"), "mean", "min", "max"])

    if sum_aggregation:
        df = df.group_by([column_uid, "week"]).agg(
            [
                pl.col("mean").sum().alias("mean"),
                pl.col("min").sum().alias("min"),
                pl.col("max").sum().alias("max"),
            ]
        )
    else:
        df = df.group_by([column_uid, "week"]).agg(
            [
                pl.col("mean").mean().alias("mean"),
                pl.col("min").min().alias("min"),
                pl.col("max").max().alias("max"),
            ]
        )

    # sort per date since dhis2 period format is "2012W9", we need to extract year and week number
    # from the period string and cast them to int before sorting, else "2012W9" will be superior to
    # "2012W32"
    df = df.sort(
        by=[
            pl.col("week").str.split("W").list.get(0).cast(int),
            pl.col("week").str.split("W").list.get(1).cast(int),
            pl.col(column_uid),
        ]
    )

    return df


def aggregate_per_month(daily: pl.DataFrame, column_uid: str, sum_aggregation: bool = False) -> pl.DataFrame:
    """Aggregate daily data per month.

    Parameters
    ----------
    daily : pl.DataFrame
        Daily data with a "month", "mean", "min", and "max" columns
        Length of the dataframe should be (n_boundaries * n_days).
    column_uid : str
        Column containing the boundary ID.
    sum_aggregation : bool, optional
        If True, sum values instead of computing the mean, for example for total precipitation data.

    Returns
    -------
    pl.DataFrame
        Monthly aggregated data of length (n_boundaries * n_months).
    """
    df = daily.select([column_uid, "month", "mean", "min", "max"])

    if sum_aggregation:
        df = df.group_by([column_uid, "month"]).agg(
            [
                pl.col("mean").sum().alias("mean"),
                pl.col("min").sum().alias("min"),
                pl.col("max").sum().alias("max"),
            ]
        )
    else:
        df = df.group_by([column_uid, "month"]).agg(
            [
                pl.col("mean").mean().alias("mean"),
                pl.col("min").min().alias("min"),
                pl.col("max").max().alias("max"),
            ]
        )

    df = df.sort(
        by=[
            pl.col("month").cast(int),
            pl.col(column_uid),
        ]
    )

    return df



"""A set of opinionated functions to extract DHIS2 metadata & data values into dataframes."""

import logging
from datetime import datetime
from typing import Literal

import polars as pl

from openhexa.toolbox.dhis2 import DHIS2

logger = logging.getLogger(__name__)


class MissingParameterError(Exception):
    """Exception raised when a required parameter is missing."""


class MissingColumnError(Exception):
    """Exception raised when a required column is missing."""


class InvalidDataTypeError(Exception):
    """Exception raised when a column has an invalid data type."""


class InvalidParameterError(Exception):
    """Exception raised when a parameter is invalid."""


DHIS2_DATE_FORMAT = "%Y-%m-%dT%H:%M:%S.%3f%z"


def get_datasets(dhis2: DHIS2, filters: list[str] | None = None) -> pl.DataFrame:
    """Extract datasets metadata.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    filters : list[str], optional
        DHIS2 query filter expressions.

    Returns
    -------
    pl.DataFrame
        Dataframe containing datasets metadata with the following columns: id, name,
        organisation_units, data_elements, indicators, period_type.
    """
    meta = dhis2.meta.datasets(
        fields="id,name,organisationUnits,dataSetElements,indicators,periodType,lastUpdated",
        filters=filters,
    )

    schema = {
        "id": str,
        "name": str,
        "organisation_units": list[str],
        "data_elements": list[str],
        "indicators": list[str],
        "periodType": str,
    }

    df = pl.DataFrame(data=meta, schema=schema, strict=True).select(
        "id",
        "name",
        "organisation_units",
        "data_elements",
        "indicators",
        pl.col("periodType").alias("period_type"),
    )

    return df.sort(by="name")


def get_data_elements(dhis2: DHIS2, filters: list[str] | None = None) -> pl.DataFrame:
    """Extract data elements metadata.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    filters : list[str], optional
        DHIS2 query filter expressions.

    Returns
    -------
    pl.DataFrame
        Dataframe containing data elements metadata with the following columns: id, name, value_type.
    """
    meta = dhis2.meta.data_elements(fields="id,name,valueType", filters=filters)
    schema = {"id": str, "name": str, "valueType": str}
    df = pl.DataFrame(meta, schema=schema)
    return df.select("id", "name", pl.col("valueType").alias("value_type"))


def get_data_element_groups(dhis2: DHIS2, filters: list[str] | None = None) -> pl.DataFrame:
    """Extract data element groups metadata.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    filters : list[str], optional
        DHIS2 query filter expressions.

    Returns
    -------
    pl.DataFrame
        Dataframe containing data element groups metadata with the following columns: id, name,
        data_elements.
    """
    meta = dhis2.meta.data_element_groups(fields="id,name,dataElements", filters=filters)
    schema = {"id": str, "name": str, "dataElements": list[str]}
    df = pl.DataFrame(meta, schema=schema)
    df = df.select("id", "name", pl.col("dataElements").alias("data_elements"))
    return df.sort(by="name")


def get_indicators(dhis2: DHIS2, filters: list[str] | None = None) -> pl.DataFrame:
    """Extract indicators metadata.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    filters : list[str], optional
        DHIS2 query filter expressions.

    Returns
    -------
    pl.DataFrame
        Dataframe containing indicators metadata with the following columns: id, name, value_type.
    """
    meta = dhis2.meta.indicators(fields="id,name,numerator,denominator", filters=filters)
    schema = {"id": str, "name": str, "numerator": str, "denominator": str}
    df = pl.DataFrame(meta, schema=schema)
    return df.select("id", "name", "numerator", "denominator")


def get_indicator_groups(dhis2: DHIS2, filters: list[str] | None = None) -> pl.DataFrame:
    """Extract indicator groups metadata.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    filters : list[str], optional
        DHIS2 query filter expressions.

    Returns
    -------
    pl.DataFrame
        Dataframe containing indicator groups metadata with the following columns: id, name,
        data_elements.
    """
    meta = dhis2.meta.indicator_groups(fields="id,name,indicators", filters=filters)
    schema = {"id": str, "name": str, "indicators": list[str]}
    df = pl.DataFrame(meta, schema=schema)
    df = df.select("id", "name", pl.col("indicators").alias("indicators"))
    return df.sort(by="name")


def get_organisation_unit_levels(dhis2: DHIS2) -> pl.DataFrame:
    """Extract organisation unit levels metadata.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.

    Returns
    -------
    pl.DataFrame
        Dataframe containing organisation unit levels metadata with the following columns: id, name,
        level.
    """
    meta = dhis2.meta.organisation_unit_levels()
    schema = {"id": str, "name": str, "level": int}
    return pl.DataFrame(data=meta, schema=schema)


def get_organisation_units(
    dhis2: DHIS2, max_level: int | None = None, filters: list[str] | None = None
) -> pl.DataFrame:
    """Extract organisation units metadata.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    max_level : int, optional
        Maximum level of organisation units to extract. If None, all levels are extracted.
    filters : list[str], optional
        DHIS2 query filter expressions.

    Returns
    -------
    pl.DataFrame
        Dataframe containing organisation units metadata with the following columns: id, name,
        level, level_{level}_id, level_{level}_name, geometry.

    Raises
    ------
    InvalidParameter
        If max_level is greater than the maximum level of the organisation units.
    """
    levels = get_organisation_unit_levels(dhis2)
    if max_level:
        if max_level > levels["level"].max():
            msg = f"max_level cannot be greater than {levels['level'].max()}"
            logger.error(msg)
            raise InvalidParameterError(msg)
        level_filter = f"level:le:{max_level}"
        if filters:
            filters = [*filters, max_level]
        else:
            filters = [level_filter]

    meta = dhis2.meta.organisation_units(fields="id,name,level,path,openingDate,closedDate,geometry", filters=filters)

    schema = {
        "id": str,
        "name": str,
        "level": int,
        "path": str,
        "openingDate": str,
        "closedDate": str,
        "geometry": str,
    }
    df = pl.DataFrame(data=meta, schema=schema)

    for row in levels.iter_rows(named=True):
        lvl = row["level"]
        if max_level:
            if lvl > max_level:
                continue

        df = df.with_columns(
            pl.col("path").str.split("/").list.slice(1).list.get(lvl - 1, null_on_oob=True).alias(f"level_{lvl}_id")
        )

        df = df.join(
            other=df.select("id", pl.col("name").alias(f"level_{lvl}_name")),
            left_on=f"level_{lvl}_id",
            right_on="id",
            how="left",
        )

    df = df.select(
        "id",
        "name",
        "level",
        pl.col("openingDate").str.to_datetime("%Y-%m-%dT%H:%M:%S.%3f").alias("opening_date"),
        pl.col("closedDate").str.to_datetime("%Y-%m-%dT%H:%M:%S.%3f").alias("closed_date"),
        *[col for col in df.columns if col.startswith("level_")],
        "geometry",
    )

    return df.sort(by=["level", "name"], descending=False)


def get_organisation_unit_groups(dhis2: DHIS2, filters: list[str] | None = None) -> pl.DataFrame:
    """Extract organisation unit groups metadata.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    filters : list[str], optional
        DHIS2 query filter expressions.

    Returns
    -------
    pl.DataFrame
        Dataframe containing organisation unit groups metadata with the following columns: id, name,
        organisation_units.
    """
    meta = dhis2.meta.organisation_unit_groups(fields="id,name,organisationUnits", filters=filters)
    schema = {"id": str, "name": str, "organisationUnits": list[str]}
    df = pl.DataFrame(meta, schema=schema)
    df = df.select("id", "name", pl.col("organisationUnits").alias("organisation_units"))
    return df.sort(by="name")


def get_category_option_combos(dhis2: DHIS2, filters: list[str] | None = None) -> pl.DataFrame:
    """Extract category option combos metadata.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    filters : str, optional
        DHIS2 query filter expression.

    Returns
    -------
    pl.DataFrame
        Dataframe containing category option combos metadata with the following columns: id, name.
    """
    meta = dhis2.meta.category_option_combos(filters=filters)
    schema = {"id": str, "name": str}
    df = pl.DataFrame(meta, schema=schema)
    return df.sort(by="name")


def get_attributes(dhis2: DHIS2, filters: list[str] | None = None) -> pl.DataFrame:
    """Extract attributes metadata.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    filters : list[str], optional
        DHIS2 query filter expressions.

    Returns
    -------
    pl.DataFrame
        Dataframe containing attributes metadata with the following columns: id, name.
    """
    r = dhis2.api.get("attributes", params={"paging": False, "fields": "id,name", "filter": filters})
    rows = r["attributes"]
    schema = {"id": str, "name": str}
    df = pl.DataFrame(rows, schema=schema)
    return df.sort(by="name")


def _data_values_to_dataframe(values: list[dict]) -> pl.DataFrame:
    """Convert a list of raw DHIS2 data values to a Polars DataFrame.

    Parameters
    ----------
    values : list[dict]
        List of data values.

    Returns
    -------
    pl.DataFrame
        Dataframe containing data values with the following columns: data_element_id, period,
        organisation_unit, category_option_combo, attribute_option_combo, value, created,
        last_updated.
    """
    schema = {
        "dataElement": str,
        "period": str,
        "orgUnit": str,
        "categoryOptionCombo": str,
        "attributeOptionCombo": str,
        "value": str,
        "created": str,
        "lastUpdated": str,
    }

    df = pl.DataFrame(data=values, schema=schema)
    df = df.select(
        pl.col("dataElement").alias("data_element_id"),
        pl.col("period"),
        pl.col("orgUnit").alias("organisation_unit_id"),
        pl.col("categoryOptionCombo").alias("category_option_combo_id"),
        pl.col("attributeOptionCombo").alias("attribute_option_combo_id"),
        pl.col("value"),
        pl.col("created").str.to_datetime(DHIS2_DATE_FORMAT).alias("created"),
        pl.col("lastUpdated").str.to_datetime(DHIS2_DATE_FORMAT).alias("last_updated"),
    )

    return df


def extract_dataset(
    dhis2: DHIS2,
    dataset: str,
    start_date: datetime | None = None,
    end_date: datetime | None = None,
    periods: list[str] | None = None,
    org_units: list[str] = None,
    org_unit_groups: list[str] = None,
    include_children: bool = False,
    last_updated: datetime | None = None,
) -> pl.DataFrame:
    """Extract dataset data values.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    dataset : str
        Dataset ID.
    start_date : str, optional
        Start date in the format "YYYY-MM-DD".
        Use either start_date and end_date or periods.
    end_date : str, optional
        End date in the format "YYYY-MM-DD".
        Use either start_date and end_date or periods.
    periods : list[str], optional
        Periods to extract data values for (ex: ["202101", "202102", "202103"]). Periods must be
        provided in DHIS2 period format. Use either start_date and end_date or periods.
    org_units : list[str], optional
        Organisation units IDs.
    org_unit_groups : list[str], optional
        Organisation unit groups IDs.
    include_children : bool, optional
        Include children organisation units.
    last_updated : datetime, optional
        Extract only data values that have been updated since this date.

    Returns
    -------
    pl.DataFrame
        Dataframe containing data values with the following columns: data_element_id, period,
        organisation_unit, category_option_combo, attribute_option_combo, value, created,
        last_updated.

    Raises
    ------
    MissingParameter
        If org_units or org_unit_groups is not provided.
    InvalidParameter
        If org_units and org_unit_groups are provided at the same time.
    """
    if org_units is None and org_unit_groups is None:
        msg = "org_units or org_unit_groups must be provided"
        logger.error(msg)
        raise MissingParameterError(msg)

    if org_units is not None and org_unit_groups is not None:
        msg = "org_units and org_unit_groups cannot be provided at the same time"
        logger.error(msg)
        raise InvalidParameterError(msg)

    if not (start_date and end_date) and not periods:
        msg = "Either start_date and end_date or periods must be provided"
        logger.error(msg)
        raise MissingParameterError(msg)

    if (start_date or end_date) and periods:
        msg = "Either start_date and end_date or periods must be provided, not both"
        logger.error(msg)
        raise InvalidParameterError(msg)

    if start_date:
        start_date = start_date.strftime("%Y-%m-%d")
    if end_date:
        end_date = end_date.strftime("%Y-%m-%d")

    values = dhis2.data_value_sets.get(
        datasets=[dataset],
        periods=periods if periods else None,
        start_date=start_date if start_date else None,
        end_date=end_date if end_date else None,
        org_units=org_units if org_units else None,
        org_unit_groups=org_unit_groups if org_unit_groups else None,
        children=include_children,
        last_updated=last_updated.isoformat() if last_updated else None,
    )

    df = _data_values_to_dataframe(values)
    return df


def extract_data_element_group(
    dhis2: DHIS2,
    data_element_group: str,
    start_date: datetime | None = None,
    end_date: datetime | None = None,
    periods: list[str] | None = None,
    org_units: list[str] = None,
    org_unit_groups: list[str] = None,
    include_children: bool = False,
    last_updated: datetime | None = None,
) -> pl.DataFrame:
    """Extract data element group data values.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    data_element_group : str
        Data element group ID.
    start_date : str, optional
        Start date in the format "YYYY-MM-DD".
        Use either start_date and end_date or periods.
    end_date : str, optional
        End date in the format "YYYY-MM-DD".
        Use either start_date and end_date or periods.
    periods : list[str], optional
        Periods to extract data values for (ex: ["202101", "202102", "202103"]). Periods must be
        provided in DHIS2 period format. Use either start_date and end_date or periods.
    org_units : list[str], optional
        Organisation units IDs.
    org_unit_groups : list[str], optional
        Organisation unit groups IDs.
    include_children : bool, optional
        Include children organisation units.
    last_updated : datetime, optional
        Extract only data values that have been updated since this date.

    Returns
    -------
    pl.DataFrame
        Dataframe containing data values with the following columns: data_element_id, period,
        organisation_unit, category_option_combo, attribute_option_combo, value, created,
        last_updated.

    Raises
    ------
    MissingParameter
        If org_units or org_unit_groups is not provided.
    InvalidParameter
        If org_units and org_unit_groups are provided at the same time.
    """
    if org_units is None and org_unit_groups is None:
        msg = "org_units or org_unit_groups must be provided"
        logger.error(msg)
        raise MissingParameterError(msg)

    if org_units is not None and org_unit_groups is not None:
        msg = "org_units and org_unit_groups cannot be provided at the same time"
        logger.error(msg)
        raise InvalidParameterError(msg)

    if not (start_date and end_date) and not periods:
        msg = "Either start_date and end_date or periods must be provided"
        logger.error(msg)
        raise MissingParameterError(msg)

    if (start_date or end_date) and periods:
        msg = "Either start_date and end_date or periods must be provided, not both"
        logger.error(msg)
        raise InvalidParameterError(msg)

    if start_date:
        start_date = start_date.strftime("%Y-%m-%d")
    if end_date:
        end_date = end_date.strftime("%Y-%m-%d")

    values = dhis2.data_value_sets.get(
        data_element_groups=[data_element_group],
        periods=periods if periods else None,
        start_date=start_date if start_date else None,
        end_date=end_date if end_date else None,
        org_units=org_units if org_units else None,
        org_unit_groups=org_unit_groups if org_unit_groups else None,
        children=include_children,
        last_updated=last_updated.isoformat() if last_updated else None,
    )

    df = _data_values_to_dataframe(values)
    return df


def extract_data_elements(
    dhis2: DHIS2,
    data_elements: list[str],
    start_date: datetime | None = None,
    end_date: datetime | None = None,
    periods: list[str] | None = None,
    org_units: list[str] = None,
    org_unit_groups: list[str] = None,
    include_children: bool = False,
    last_updated: datetime | None = None,
) -> pl.DataFrame:
    """Extract data elements data values.

    Only DHIS2 versions >= 2.39 are supported.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    data_elements : list[str]
        Data elements IDs.
    start_date : str, optional
        Start date in the format "YYYY-MM-DD".
        Use either start_date and end_date or periods.
    end_date : str, optional
        End date in the format "YYYY-MM-DD".
        Use either start_date and end_date or periods.
    periods : list[str], optional
        Periods to extract data values for (ex: ["202101", "202102", "202103"]). Periods must be
        provided in DHIS2 period format. Use either start_date and end_date or periods.
    org_units : list[str], optional
        Organisation units IDs.
    org_unit_groups : list[str], optional
        Organisation unit groups IDs.
    include_children : bool, optional
        Include children organisation units.
    last_updated : datetime, optional
        Extract only data values that have been updated since this date.

    Returns
    -------
    pl.DataFrame
        Dataframe containing data values with the following columns: data_element_id, period,
        organisation_unit, category_option_combo, attribute_option_combo, value, created,
        last_updated.

    Raises
    ------
    MissingParameter
        If org_units or org_unit_groups is not provided.
    InvalidParameter
        If org_units and org_unit_groups are provided at the same time.
    """
    if org_units is None and org_unit_groups is None:
        msg = "org_units or org_unit_groups must be provided"
        logger.error(msg)
        raise MissingParameterError(msg)

    if org_units is not None and org_unit_groups is not None:
        msg = "org_units and org_unit_groups cannot be provided at the same time"
        logger.error(msg)
        raise InvalidParameterError(msg)

    if not (start_date and end_date) and not periods:
        msg = "Either start_date and end_date or periods must be provided"
        logger.error(msg)
        raise MissingParameterError(msg)

    if (start_date or end_date) and periods:
        msg = "Either start_date and end_date or periods must be provided, not both"
        logger.error(msg)
        raise InvalidParameterError(msg)

    if start_date:
        start_date = start_date.strftime("%Y-%m-%d")
    if end_date:
        end_date = end_date.strftime("%Y-%m-%d")

    values = dhis2.data_value_sets.get(
        data_elements=data_elements,
        periods=periods if periods else None,
        start_date=start_date if start_date else None,
        end_date=end_date if end_date else None,
        org_units=org_units if org_units else None,
        org_unit_groups=org_unit_groups if org_unit_groups else None,
        children=include_children,
        last_updated=last_updated.isoformat() if last_updated else None,
    )

    df = _data_values_to_dataframe(values)
    return df


def extract_analytics(
    dhis2: DHIS2,
    periods: list[str],
    data_elements: list[str] = None,
    data_element_groups: list[str] = None,
    indicators: list[str] = None,
    indicator_groups: list[str] = None,
    org_units: list[str] = None,
    org_unit_groups: list[str] = None,
    org_unit_levels: list[int] = None,
) -> pl.DataFrame:
    """Extract aggregated data values using an Analytics query.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    periods : list[str]
        Periods to extract data values for (ex: ["202101", "202102", "202103"]). Periods must be
        provided in DHIS2 period format.
    data_elements : list[str], optional
        Data elements IDs.
    data_element_groups : list[str], optional
        Data element groups IDs.
    indicators : list[str], optional
        Indicators IDs.
    indicator_groups : list[str], optional
        Indicator groups IDs.
    org_units : list[str], optional
        Organisation units IDs.
    org_unit_groups : list[str], optional
        Organisation unit groups IDs.
    org_unit_levels : list[int], optional
        Organisation unit levels.

    Returns
    -------
    pl.DataFrame
        Dataframe containing data values with the following columns: data_element_id or
        indicator_id, category_option_combo_id, organisation_unit_id, period, value.
    """
    # always include COCs by default, except when extracting indicators
    include_cocs = False if indicators or indicator_groups else True

    if (data_elements or data_element_groups) and (indicators or indicator_groups):
        msg = "Data elements and indicators cannot be requested at the same time"
        logger.error(msg)
        raise ValueError(msg)

    values = dhis2.analytics.get(
        data_elements=data_elements,
        data_element_groups=data_element_groups,
        indicators=indicators,
        indicator_groups=indicator_groups,
        periods=periods,
        org_units=org_units,
        org_unit_groups=org_unit_groups,
        org_unit_levels=org_unit_levels,
        include_cocs=include_cocs,
    )

    schema = {"dx": str, "ou": str, "pe": str, "value": str}

    if include_cocs:
        schema["co"] = str

    df = pl.DataFrame(data=values, schema=schema)

    # dx is either data_element_id or indicator_id, depending on the analytics query
    dx = "data_element_id" if data_elements or data_element_groups else "indicator_id"

    if include_cocs:
        return df.select(
            pl.col("dx").alias(dx),
            pl.col("co").alias("category_option_combo_id"),
            pl.col("ou").alias("organisation_unit_id"),
            pl.col("pe").alias("period"),
            pl.col("value"),
        )
    return df.select(
        pl.col("dx").alias(dx),
        pl.col("ou").alias("organisation_unit_id"),
        pl.col("pe").alias("period"),
        pl.col("value"),
    )


def _validate_data_values(df: pl.DataFrame) -> None:
    """Validate data values dataframe for import into DHIS2.

    Parameters
    ----------
    df : pl.DataFrame
        Dataframe containing data values.

    Raises
    ------
    MissingColumn
        If a required column is missing.
    InvalidDataType
        If a column has an invalid data type.
    """
    if "data_element_id" not in df.columns:
        msg = "Missing data_element_id column"
        logger.error(msg)
        raise MissingColumnError(msg)
    if "organisation_unit_id" not in df.columns:
        msg = "Missing organisation_unit_id column"
        logger.error(msg)
        raise MissingColumnError(msg)
    if "period" not in df.columns:
        msg = "Missing period column"
        logger.error(msg)
        raise MissingColumnError(msg)
    if "category_option_combo_id" not in df.columns:
        msg = "Missing category_option_combo_id column"
        logger.error(msg)
        raise MissingColumnError(msg)
    if "attribute_option_combo_id" not in df.columns:
        msg = "Missing attribute_option_combo_id column"
        logger.error(msg)
        raise MissingColumnError(msg)
    if "value" not in df.columns:
        msg = "Missing value column"
        logger.error(msg)
        raise MissingColumnError(msg)
    if df["data_element_id"].dtype != pl.Utf8:
        msg = "data_element_id must be of type String"
        logger.error(msg)
        raise InvalidDataTypeError(msg)
    if df["organisation_unit_id"].dtype != pl.Utf8:
        msg = "organisation_unit_id must be of type String"
        logger.error(msg)
        raise InvalidDataTypeError(msg)
    if df["period"].dtype != pl.Utf8:
        msg = "period must be of type String"
        logger.error(msg)
        raise InvalidDataTypeError(msg)
    if df["category_option_combo_id"].dtype != pl.Utf8:
        msg = "category_option_combo_id must be of type String"
        logger.error(msg)
        raise InvalidDataTypeError(msg)
    if df["attribute_option_combo_id"].dtype != pl.Utf8:
        msg = "attribute_option_combo_id must be of type String"
        logger.error(msg)
        raise InvalidDataTypeError(msg)
    if df["value"].dtype != pl.Utf8:
        msg = "value must be of type String"
        logger.error(msg)
        raise InvalidDataTypeError(msg)


def _map_uids(df: pl.DataFrame, **mappings) -> pl.DataFrame:
    """Replace UIDs in a dataframe based on mappings.

    Replacements are strict, which means that if a uid is not found in the mapping,
    it will be assigned to None.

    Parameters
    ----------
    df : pl.DataFrame
        Dataframe containing data values.
    mappings : dict
        Mappings to replace UIDs, with column name as key and dict mapping as value.

    Returns
    -------
    pl.DataFrame
        Dataframe with UIDs replaced based on mappings.
    """
    for col, mapping in mappings.items():
        if mapping:
            if col not in df.columns:
                msg = f"Missing {col} column"
                logger.error(msg)
                raise MissingColumnError(msg)
            df = df.with_columns(pl.col(col).replace_strict(mapping, default=None))
    return df


def import_data_values(
    dhis2: DHIS2,
    data: pl.DataFrame,
    org_units_mapping: dict | None = None,
    data_elements_mapping: dict | None = None,
    category_option_combos_mapping: dict | None = None,
    attribute_option_combos_mapping: dict | None = None,
    import_strategy: Literal["CREATE", "UPDATE", "CREATE_AND_UPDATE", "DELETE"] = "CREATE",
    dry_run: bool = True,
) -> dict:
    """Import data values to a DHIS2 instance.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    data : pl.DataFrame
        Dataframe containing data values with the following columns: data_element_id, period,
        organisation_unit_id, category_option_combo_id, attribute_option_combo_id, value.
    org_units_mapping : dict, optional
        Organisation units mapping with old UID as key and new UID as value.
    data_elements_mapping : dict, optional
        Data elements mapping with old UID as key and new UID as value.
    category_option_combos_mapping : dict, optional
        Category option combos mapping with old UID as key and new UID as value.
    attribute_option_combos_mapping : dict, optional
        Attribute option combos mapping with old UID as key and new UID as value.
    import_strategy : str, optional
        Import strategy. One of "CREATE", "UPDATE", "CREATE_AND_UPDATE", "DELETE". Default is "CREATE".
    dry_run : bool, optional
        Perform a dry run. Default is True.

    Returns
    -------
    dict
        Import report.

    Raises
    ------
    MissingColumn
        If a required column is missing.
    InvalidDataType
        If a column has an invalid data type.
    """
    data = _map_uids(
        df=data,
        organisation_unit_id=org_units_mapping,
        data_element_id=data_elements_mapping,
        category_option_combo_id=category_option_combos_mapping,
        attribute_option_combo_id=attribute_option_combos_mapping,
    )

    # ignore rows with null values
    nrows = len(data)
    data = data.drop_nulls()
    nrows_dropped = nrows - len(data)
    logger.debug(f"Dropped {nrows_dropped} rows with null values")

    _validate_data_values(df=data)

    data_values = data.select(
        pl.col("data_element_id").alias("dataElement"),
        pl.col("period"),
        pl.col("organisation_unit_id").alias("orgUnit"),
        pl.col("category_option_combo_id").alias("categoryOptionCombo"),
        pl.col("attribute_option_combo_id").alias("attributeOptionCombo"),
        pl.col("value"),
    ).to_dicts()

    report = dhis2.data_value_sets.post(
        data_values=data_values,
        import_strategy=import_strategy,
        dry_run=dry_run,
        skip_validation=True,
    )

    return report


def _extract_attribute_values(attribute_values: list[dict], mapping: dict) -> list[dict]:
    """Reshape attributeValue struct returned by the DHIS2 API and join attribute names.

    Parameters
    ----------
    attribute_values : list[dict]
        List of attribute values.
    mapping : dict
        Mapping of attribute IDs to attribute names.

    Returns
    -------
    list[dict]
        List of attribute values with the following keys: attribute_id, attribute_name, value
    """
    if attribute_values is None:
        return None

    values = []

    for attribute in attribute_values:
        attribute_id = attribute["attribute"]["id"]
        attribute_name = mapping[attribute_id]

        if attribute_id == "coordinates":
            continue

        values.append(
            {
                "attribute_id": attribute_id,
                "attribute_name": attribute_name,
                "value": attribute["value"],
            }
        )

    return values


def extract_organisation_unit_attributes(dhis2: DHIS2) -> pl.DataFrame:
    """Extract organisation unit attributes.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.

    Returns
    -------
    pl.DataFrame
        Dataframe containing organisation unit attributes with the following columns:
        organisation_unit_id, organisation_unit_name, attribute_id, attribute_name, value.
    """
    attributes = get_attributes(dhis2)
    mapping = {attr["id"]: attr["name"] for attr in attributes.to_dicts()}
    org_units = dhis2.meta.organisation_units(fields="id,name,attributeValues")
    schema = {
        "id": str,
        "name": str,
        "attributeValues": pl.List(pl.Struct({"attribute": pl.Struct({"id": str}), "value": str})),
    }
    org_units = pl.DataFrame(org_units, schema=schema)

    rows = []
    for row in org_units.iter_rows(named=True):
        attrs = _extract_attribute_values(row["attributeValues"], mapping)
        for attr in attrs:
            rows.append(
                {
                    "organisation_unit_id": row["id"],
                    "organisation_unit_name": row["name"],
                    "attribute_id": attr["attribute_id"],
                    "attribute_name": attr["attribute_name"],
                    "value": attr["value"],
                }
            )

    schema = {
        "organisation_unit_id": str,
        "organisation_unit_name": str,
        "attribute_id": str,
        "attribute_name": str,
        "value": str,
    }

    return pl.DataFrame(rows, schema=schema)


def get_programs(dhis2: DHIS2, filters: list[str] | None = None) -> pl.DataFrame:
    """Extract programs metadata.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    filters : list[str], optional
        DHIS2 query filter expressions.

    Returns
    -------
    pl.DataFrame
        Dataframe containing programs metadata with the following columns: id, name.
    """
    rows = []
    for page in dhis2.api.get_paged("programs", params={"fields": "id,name", "filter": filters}):
        rows.extend(page["programs"])

    schema = {"id": str, "name": str}
    return pl.DataFrame(rows, schema=schema)


def get_tracked_entity_types(dhis2: DHIS2, filters: list[str] | None = None) -> pl.DataFrame:
    """Extract tracked entity types metadata.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    filters : list[str], optional
        DHIS2 query filter expressions.

    Returns
    -------
    pl.DataFrame
        Dataframe containing tracked entity types metadata with the following columns: id, name.
    """
    rows = []
    for page in dhis2.api.get_paged("trackedEntityTypes", params={"fields": "id,name", "filter": filters}):
        rows.extend(page["trackedEntityTypes"])

    schema = {"id": str, "name": str}
    return pl.DataFrame(rows, schema=schema)


def extract_events(
    dhis2: DHIS2,
    program_id: str,
    org_unit_parents: list[str],
    occurred_after: str | None = None,
    occurred_before: str | None = None,
) -> pl.DataFrame:
    """Extract events data.

    Parameters
    ----------
    dhis2 : DHIS2
        DHIS2 instance.
    program_id : str
        Program UID.
    org_unit_parents : list[str]
        Organisation unit parents UIDs. Event data will be extracted for all descendants of these
        organisation units.
    occurred_after : str, optional
        Start date in the format "YYYY-MM-DD".
    occurred_before : str, optional
        End date in the format "YYYY-MM-DD".

    Returns
    -------
    pl.DataFrame
        Dataframe containing events data with the following columns: event_id, program_id,
        organisation_unit_id, tracked_entity_instance_id, event_date.
    """
    data = []
    for org_unit in org_unit_parents:
        params = {
            "orgUnit": org_unit,
            "program": program_id,
            "ouMode": "DESCENDANTS",
            "fields": "event,status,program,programStage,orgUnit,occurredAt,deleted,attributeOptionCombo,dataValues",
            "totalPages": True,
        }
        if occurred_after:
            params["occurredAfter"] = occurred_after
        if occurred_before:
            params["occurredBefore"] = occurred_before
        for page in dhis2.api.get_paged("tracker/events", params=params):
            data.extend(page["instances"])

    schema = {
        "event": str,
        "status": str,
        "program": str,
        "programStage": str,
        "orgUnit": str,
        "occurredAt": str,
        "deleted": bool,
        "attributeOptionCombo": str,
        "dataValues": pl.List(pl.Struct({"dataElement": str, "value": str})),
    }

    df = pl.DataFrame(data, schema=schema)
    df = df.select(
        [
            pl.col("event").alias("event_id"),
            pl.col("status"),
            pl.col("program").alias("program_id"),
            pl.col("programStage").alias("program_stage_id"),
            pl.col("orgUnit").alias("organisation_unit_id"),
            pl.col("occurredAt").str.to_datetime("%Y-%m-%dT%H:%M:%S.%3f").alias("occurred_at"),
            pl.col("deleted"),
            pl.col("attributeOptionCombo").alias("attribute_option_combo_id"),
            pl.col("dataValues"),
        ]
    )

    # build a new dataframe with one row per event data value, instead of storing all data values
    # in a column of type list[struct]

    new_rows = []

    for row in df.iter_rows(named=True):
        for data_value in row["dataValues"]:
            new_row = {col: value for col, value in row.items() if col != "dataValues"}
            new_row["data_element_id"] = data_value["dataElement"]
            new_row["value"] = data_value["value"]
            new_rows.append(new_row)

    schema = {
        "event_id": str,
        "status": str,
        "program_id": str,
        "program_stage_id": str,
        "organisation_unit_id": str,
        "occurred_at": pl.Datetime(time_unit="ms", time_zone="UTC"),
        "deleted": bool,
        "attribute_option_combo_id": str,
        "data_element_id": str,
        "value": str,
    }

    return pl.DataFrame(new_rows, schema=schema)


def join_object_names(
    df: pl.DataFrame,
    data_elements: pl.DataFrame | None = None,
    indicators: pl.DataFrame | None = None,
    organisation_units: pl.DataFrame | None = None,
    category_option_combos: pl.DataFrame | None = None,
) -> pl.DataFrame:
    if (
        (data_elements is None)
        and (organisation_units is None)
        and (category_option_combos is None)
        and (indicators is None)
    ):
        msg = "At least one of data_elements, organisation_units or category_option_combos must be provided"
        logger.error(msg)
        raise ValueError(msg)

    if data_elements is not None and "data_element_name" not in df.columns:
        df = df.join(
            other=data_elements.select("id", pl.col("name").alias("data_element_name")),
            left_on="data_element_id",
            right_on="id",
            how="left",
        )

    if indicators is not None and "indicator_name" not in df.columns:
        df = df.join(
            other=indicators.select("id", pl.col("name").alias("indicator_name")),
            left_on="indicator_id",
            right_on="id",
            how="left",
        )

    if organisation_units is not None and "organisation_unit_name" not in df.columns:
        ou_ids = [col for col in organisation_units.columns if col.startswith("level_") and col.endswith("_id")]
        ou_names = [col for col in organisation_units.columns if col.startswith("level_") and col.endswith("_name")]
        df = df.join(
            other=organisation_units.select("id", *ou_ids, *ou_names),
            left_on="organisation_unit_id",
            right_on="id",
            how="left",
        )

    if category_option_combos is not None and "category_option_combo_name" not in df.columns:
        df = df.join(
            other=category_option_combos.select("id", pl.col("name").alias("category_option_combo_name")),
            left_on="category_option_combo_id",
            right_on="id",
            how="left",
        )

    COLUMNS = [
        "dataset_name",
        "dataset_id",
        "dataset_period_type",
        "data_element_id",
        "data_element_name",
        "indicator_id",
        "indicator_name",
        "organisation_unit_id",
        "organisation_unit_name",
        "category_option_combo_id",
        "category_option_combo_name",
        "attribute_option_combo_id",
        "period",
        "value",
        *[col for col in df.columns if col.startswith("level_")],
        "created",
        "last_updated",
    ]

    # if additional columns were present in the original dataframe, keep them at the end
    COLUMNS += [col for col in df.columns if col not in COLUMNS]

    return df.select([col for col in COLUMNS if col in df.columns])



"""KoboToolbox API."""

from __future__ import annotations

from functools import cached_property
from typing import List

import requests


class Field:
    def __init__(self, meta: dict):
        """Initialize a Field object from json metadata."""
        self.meta = meta

    def __repr__(self) -> str:
        return f'Field("{self.name}")'

    @property
    def uid(self) -> str:
        """Get field UID."""
        return self.meta.get("$kuid")

    @property
    def name(self) -> str:
        """Get field name."""
        return self.meta.get("name", "")

    @property
    def xpath(self) -> str:
        """Get field xpath."""
        return self.meta.get("$xpath")

    @property
    def label(self) -> str:
        """Get field full label."""
        if "label" in self.meta:
            return self.meta["label"][0]
        else:
            return None

    @property
    def type(self) -> str:
        """Get field data type."""
        return self.meta.get("type")

    @property
    def list_name(self) -> str:
        """Get the list of available choices for the field."""
        return self.meta.get("select_from_list_name")


class Survey:
    def __init__(self, client: Api, meta: dict):
        """Initialize a Survey object from json metadata."""
        self.client = client
        self.meta = meta

    def __repr__(self) -> str:
        return f'Survey("{self.name}")'

    @property
    def uid(self) -> str:
        return self.meta["uid"]

    @property
    def name(self) -> str:
        return self.meta.get("name")

    @cached_property
    def fields(self) -> List[Field]:
        """All available fields in survey."""
        fields = []
        for meta in self.meta["content"]["survey"]:
            fields.append(Field(meta))
        return fields

    def get_field(self, uid: str) -> Field:
        """Get field object from its UID."""
        for meta in self.meta["content"]["survey"]:
            if meta.get("$kuid") == uid:
                return Field(meta)

    def get_field_from_name(self, name: str) -> Field:
        """Get field object from its name."""
        for meta in self.meta["content"]["survey"]:
            if meta.get("name") == name:
                return Field(meta)

    @cached_property
    def choices(self) -> dict:
        """Get all choice lists."""
        choice_lists = {}
        if "choices" not in self.meta["content"]:
            return None
        for choice in self.meta["content"]["choices"]:
            if "list_name" in choice:
                list_name = choice["list_name"]
                if list_name not in choice_lists:
                    choice_lists[list_name] = []
                choice_lists[list_name].append(choice)
        return choice_lists

    def get_data(self) -> List[dict]:
        """Download survey data."""
        r = self.client.session.get(self.meta["data"])
        r.raise_for_status()
        return r.json().get("results")

    @cached_property
    def _xpaths_labels_mapping(self) -> dict:
        """Get a mapping of fields xpaths and labels."""
        mapping = {}
        for field in self.fields:
            xpath = field.meta.get("$xpath")
            label = field.meta.get("label")
            if xpath and label:
                mapping[xpath] = " ".join(label[0].split())
        return mapping

    def get_field_from_xpath(self, xpath: str) -> Field:
        """Get field object from its xpath."""
        for field in self.fields:
            if field.xpath == xpath:
                return field
        raise ValueError("Field not found")

    def get_label_from_xpath(self, xpath: str) -> str:
        """Get field label from its xpath."""
        for field in self.fields:
            xpath = field.meta.get("$xpath")
            if xpath == xpath:
                label = field.meta.get("label")
                if label:
                    return label[0]
                else:
                    raise ValueError("No label found for xpath")
        raise ValueError("Field not found")

    @property
    def labels(self) -> dict:
        """Get a mapping of fields xpaths and labels."""
        mapping = {}
        for field in self.fields:
            xpath = field.meta.get("$xpath")
            label = field.meta.get("label")
            if xpath and label:
                mapping[xpath] = " ".join(label[0].split())
        return mapping

    def value_from_choice(self, choice: str):
        """Get choice value from choice id."""
        choices = self.choices.get(choice)
        if not choices:
            raise ValueError("Choice list name not found")
        for c in choices:
            if c.get("name") == choice:
                label = c.get("label")
                if label:
                    return label[0]


class AuthenticationError(Exception):
    pass


class Api:
    def __init__(self, url: str):
        self.url = url.rstrip("/")
        self.session = requests.Session()

    def authenticate(self, token: str):
        self.session.headers["Authorization"] = f"Token {token}"

    def check_authentication(self):
        if "Authorization" not in self.session.headers:
            raise AuthenticationError("Not authenticated")

    def _get_assets(self) -> List[dict]:
        r = self.session.get(f"{self.url}/assets.json")
        r.raise_for_status()
        return r.json()["results"]

    @cached_property
    def surveys(self) -> List[dict]:
        """List UID and names of available surveys."""
        surveys = []
        assets = self._get_assets()
        for asset in assets:
            if asset.get("asset_type") == "survey":
                surveys.append({"uid": asset.get("uid"), "name": asset.get("name")})
        return surveys

    def get_survey(self, uid: str) -> Survey:
        """Get survey from its UID."""
        r = self.session.get(f"{self.url}/assets/{uid}.json")
        r.raise_for_status()
        return Survey(client=self, meta=r.json())

    def get_survey_data(self, survey: Survey) -> dict:
        """Get survey data."""
        r = self.session.get(survey.meta["data"])
        r.raise_for_status()
        return r.json().get("results")
